"""
Object representation of operations on TTree branches

The aim is to provide provide sufficiently flexible and complete foundation
for the development of efficient histogram-filling programs
through the use of python wrappers (see e.g. treeproxies).
"""

import binascii
import hashlib
import logging
from collections import defaultdict
from itertools import chain, combinations, repeat, zip_longest

logger = logging.getLogger(__name__)

boolType = "bool"
intType = "int"
floatType = "double"
SizeType = "std::size_t"


def equal_elements(seqA, seqB):
    return all(ia == ib for ia, ib in zip_longest(seqA, seqB))


def op_digest(*args):
    h = hashlib.blake2b(digest_size=16, salt=b"bamboo TupleOp")
    for arg in args:
        if isinstance(arg, TupleOp):
            h.update(arg.digest)
        elif isinstance(arg, str):
            h.update(arg.encode())
        elif isinstance(arg, bytes):
            h.update(arg)
        else:
            raise ValueError(f"Unsupported content: {arg!r}")
    return h.digest()


def hexStr_digest(digest):
    return binascii.b2a_hex(digest).decode()


def op_hash(*args):
    if len(args) == 1:
        return hash(args[0])
    else:
        return hash(tuple(args))


class TupleOpCache:
    __slots__ = ("hash", "repr", "digest")

    def __init__(self):
        self.hash = None
        self.repr = None
        self.digest = None

    def __bool__(self):
        return self.hash is not None or self.repr is not None


class TupleOp:
    """ Interface & base class for operations on leafs and resulting objects / values

    Instances should be defined once, and assumed immutable by all observers
    (they should only ever be modified just after construction, preferably by the owner).
    Since a value-based hash (and repr) is cached on first use, violating this rule
    might lead to serious bugs. In case of doubt the clone method can be used to
    obtain an independent copy.
    Subclasses should define a result property and clone, _repr, _eq, and optionally deps methods.
    """
    # this means all deriving classes need to define __slots__ (for performance)
    __slots__ = ("typeName", "_cache", "canDefine")

    def __init__(self, typeName):
        self.typeName = typeName
        self._cache = TupleOpCache()
        self.canDefine = True

    def clone(self, memo=None, select=lambda nd: True, selClones=None):
        """ Create an independent copy (with empty repr/hash cache) of the (sub)expression """
        if memo is None:  # top-level, construct the dictionary
            memo = dict()
        if id(self) in memo:
            return memo[id(self)]
        else:
            cp = self._clone(memo, select, selClones=selClones)
            cln = cp if cp is not None else self
            memo[id(self)] = cln
            return cln

    # simple version, call clone of attributes without worrying about memo
    def _clone(self, memo, select, selClones=None):
        """
        Implementation of clone - to be overridden by all subclasses
            (memo is dealt with by clone, so simply construct, calling
            .clone(memo=memo, select=select) on TupleOp attributes.
            Return None if no clone is needed (based on select)
        """
        if select(self):
            cln = self.__class__(self.typeName)
            if selClones is not None:
                selClones.append(cln)
            return cln

    def deps(self, defCache=None, select=(lambda x: True), includeLocal=False):
        """ Dependent TupleOps iterator """
        yield from []

    @property
    def result(self):
        """ Proxy to the result of this (sub)expression """
        from .treeproxies import makeProxy
        return makeProxy(self)

    # subclasses should define at least _clone, _repr, and _eq (value-based)
    def __repr__(self):
        """ String representation (used for hash, and lazily cached) """
        if self._cache.repr is None:
            self._cache.repr = self._repr()
        return self._cache.repr

    def _repr(self):
        """
        __repr__ implementation - to be overridden by all subclasses
            (caching is in top-level __repr__)
        """
        return f"TupleOp({self.typeName})"

    def __hash__(self):
        """ Value-based hash (lazily cached) """
        if self._cache.hash is None:
            self._cache.hash = self._hash()
        return self._cache.hash

    def _hash(self, fun=op_hash):
        """
        __hash__ and digest implementation - subclasses please override
            (caching is in top-level)
        """
        return fun(repr(self))

    @property
    def digest(self):
        """ Get a 16-byte digest (cached) """
        if self._cache.digest is None:
            self._cache.digest = self._hash(fun=op_digest)
        return self._cache.digest

    def __eq__(self, other):
        """
        Identity or value-based equality comparison
            (same object and unequal should be fast)
        """
        # _eq may end up being quite expensive, but should almost never be called
        return (id(self) == id(other)
                or (self.__hash__() == hash(other) and self.__class__ == other.__class__
                    and self._eq(other)))

    def _eq(self, other):
        """
        value-based __eq__ implementation - to be overridden by all subclasses
            (protects against hash collisions; hash and class are checked to be equal already)
        """
        return True

    def get_cppStr(self, defCache=None):
        """ Interface method: generate a C++ string from the operation

        :param defCache: cache with defined operations and symbols;
            a minimal implementation is :py:class:`~bamboo.treeoperations.CppStrRedir`
            (which does not do any caching)
        """
        pass


# implementations are split out, see treeproxies
class TupleBaseProxy:
    """
    Interface & base class for proxies
    """
    def __init__(self, op):
        self._parent = op

    @property
    def op(self):
        if self._parent is None:
            raise ValueError(f"Cannot get operation for {self!r}, abstract base class / empty parent")
        return self._parent

    @property
    def _typeName(self):
        return self._parent.typeName


class CppStrRedir:
    """ Expression cache interface. Default implementation: no caching """
    def __init__(self):
        self._iFun = 0

    def __call__(self, arg):
        return arg.get_cppStr(defCache=self)

    def symbol(self, decl):
        """
        Define (or get) a new C++ symbol for the declaration

        decl should contain the code, with <<name>> where the name should go.  Returns the unique name
        """
        raise NotImplementedError(f"Asked to add defined symbol for {decl}, but that's not supported")

    def stop(self, op):
        return False

    def getColName(self, op):
        return None

    def shouldDefine(self, arg):
        return False


cppNoRedir = CppStrRedir()


class ForwardingOp(TupleOp):
    """ Transparent wrapper (base for marking parts of the tree, e.g. things with systematic variations) """
    __slots__ = ("wrapped",)

    def __init__(self, wrapped, canDefine=None):
        self.wrapped = wrapped
        super().__init__(self.wrapped.typeName)
        self.canDefine = canDefine if canDefine is not None else self.wrapped.canDefine

    def _clone(self, memo, select, selClones=None):
        clWr = self.wrapped.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if isSel or id(clWr) != id(self.wrapped):
            cln = self.__class__(clWr, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            if select(self.wrapped):
                yield self.wrapped
            yield from self.wrapped.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    @property
    def result(self):
        wrapRes = self.wrapped.result
        wrapRes._parent = self
        return wrapRes

    def _repr(self):
        return f"{self.__class__.__name__}({self.wrapped!r})"

    def _hash(self, fun=op_hash):
        return fun(self.wrapped)

    def _eq(self, other):
        return self.wrapped == other.wrapped

    def get_cppStr(self, defCache=cppNoRedir):
        return defCache(self.wrapped)


class Const(TupleOp):
    """ Hard-coded number (or expression) """
    __slots__ = ("value")

    def __init__(self, typeName, value):
        super().__init__(typeName)
        self.value = value

    def _clone(self, memo, select, selClones=None):
        if select(self):
            cln = self.__class__(self.typeName, self.value)
            if selClones is not None:
                selClones.append(cln)
            return cln

    def _repr(self):
        return f"{self.__class__.__name__}({self.typeName!r}, {self.value!r})"

    def _hash(self, fun=op_hash):
        return fun(self._repr())

    def _eq(self, other):
        return self.typeName == other.typeName and self.value == other.value

    # backends
    def get_cppStr(self, defCache=None):
        try:
            if abs(self.value) == float("inf"):  # may raise an exception
                return "std::numeric_limits<{0}>::{mnmx}()".format(
                    self.typeName, mnmx=("lowest" if self.value < 0. else "max"))
        except Exception:
            pass
        return str(self.value)  # should maybe be type-aware...

    def getPyROOT(self, gbl):
        if self.typeName == "std::string":
            return self.value.strip('"')
        return self.value


class Parameter(Const):
    """ Constant that should not be hardcoded, to reduce the amount of generated code """
    __slots__ = tuple()

    def __init__(self, typeName, value):
        super().__init__(typeName, value)

    @property
    def name(self):  # unique and usable as value. Will be replaced by myArg* where needed
        tStr = str(self.typeName).replace(":", "_")
        return f"/* par_{tStr} */ {self.value!s} "

    def get_cppStr(self, defCache=None):
        return self.name


class GetColumn(TupleOp):
    """ Get a column value """
    __slots__ = ("typeName", "name")

    def __init__(self, typeName, name):
        super().__init__(typeName)
        self.name = name

    def _clone(self, memo, select, selClones=None):
        if select(self):
            cln = self.__class__(self.typeName, self.name)
            if selClones is not None:
                selClones.append(cln)
            return cln

    def _repr(self):
        return f"{self.__class__.__name__}({self.typeName!r}, {self.name!r})"

    def _hash(self, fun=op_hash):
        return fun(self._repr())

    def _eq(self, other):
        return self.typeName == other.typeName and self.name == other.name

    def get_cppStr(self, defCache=None):
        return self.name


class GetArrayColumn(TupleOp):
    """ Get the number from a leaf """
    __slots__ = ("valueType", "name", "length")

    def __init__(self, valueType, name, length):
        super().__init__(f"{valueType}[]")
        self.valueType = valueType
        self.name = name
        self.length = length

    def _clone(self, memo, select, selClones=None):
        lnCl = self.length.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if isSel or id(lnCl) != id(self.length):
            cln = self.__class__(self.typeName, self.name, lnCl)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            if select(self.length):
                yield self.length
            yield from self.length.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    @property
    def result(self):
        from .treeproxies import ArrayProxy
        return ArrayProxy(self, self.valueType, self.length.result)

    def _repr(self):
        return f"{self.__class__.__name__}({self.typeName!r}, {self.name!r}, {self.length!r})"

    def _hash(self, fun=op_hash):
        return fun(self._repr())

    def _eq(self, other):
        return self.typeName == other.typeName and self.name == other.name and self.length == other.length

    def get_cppStr(self, defCache=None):
        return self.name


# helper
def adaptArg(arg, typeHint=None):
    if isinstance(arg, TupleBaseProxy):
        return arg.op
    elif isinstance(arg, TupleOp):
        return arg
    elif typeHint is not None:  # make a constant
        from . import treeproxies as _tp
        if isinstance(arg, str):  # string, needs quote
            return Const(typeHint, f'"{arg}"')
        elif typeHint in _tp._boolTypes:
            return Const(typeHint, "true" if arg else "false")
        elif isinstance(arg, int):
            # TODO apply C++ rules for type of an integer literal
            return Const(typeHint, arg)
        elif isinstance(arg, float):
            return Const(typeHint, arg)
        else:
            logger.warning(f"Constructing unchecked literal of type {typeHint}: {arg}")
            return Const(typeHint, arg)
    else:
        raise ValueError("Should either get an expression, or some kind of type hint")


def _moTp_1EnsureFloat(arg):
    from . import treeproxies as _tp
    if arg.typeName in _tp._floatTypes:
        return arg.typeName
    else:
        if arg.typeName not in _tp._integerTypes:
            logger.warning(f"Unknown numeric type {arg.typeName}, promoting to {floatType}")
        return floatType


def _moTp_combMaybeFloat(*args):
    from . import treeproxies as _tp
    if any(arg.typeName in _tp._floatTypes for arg in args):
        if any(arg.typeName in ("Double_t", "double") for arg in args):
            return "double"  # 64-bit
        else:
            return "float"  # 32-bit
    else:  # TODO apply the rules for integers
        return args[0].typeName


def _moTp_minmax(a, b):
    from . import treeproxies as _tp
    if a.typeName == b.typeName:
        return a.typeName
    elif a.typeName in _tp._floatTypes or b.typeName in _tp._floatTypes:
        return floatType
    else:  # TODO apply the rules for integers
        return a.typeName


mathOpFuns_typeAndCppStr = {
    "add": (_moTp_combMaybeFloat, (lambda *args: "( {} )".format(" + ".join(args)))),
    "multiply": (_moTp_combMaybeFloat, (lambda *args: "( {} )".format(" * ".join(args)))),
    "subtract": (_moTp_combMaybeFloat, "( {} - {} )".format),
    "divide": ((lambda a1, a2: a1.typeName), "( {} / {} )".format),
    "floatdiv": (floatType, "( 1.*{} / {} )".format),
    "mod": ((lambda a1, a2: a1.typeName), "( {} % {} )".format),
    "neg": (None, "( -{} )".format),
    #
    "lt": (boolType, "( {} <  {} )".format),
    "le": (boolType, "( {} <= {} )".format),
    "eq": (boolType, "( {} == {} )".format),
    "ne": (boolType, "( {} != {} )".format),
    "gt": (boolType, "( {} >  {} )".format),
    "ge": (boolType, "( {} >= {} )".format),
    "and": (boolType, (lambda *args: "( {} )".format(" && ".join(args)))),
    "or": (boolType, (lambda *args: "( {} )".format(" || ".join(args)))),
    "not": (boolType, "( ! {} )".format),
    #
    # TODO next three: apply the rules for integer type
    "band": (intType, lambda *args: "( {} )".format(" & ".join(args))),
    "bor": (intType, lambda *args: "( {} )".format(" | ".join(args))),
    "bxor": (intType, lambda *args: "( {} )".format(" ^ ".join(args))),
    "bnot": (None, "( ~ {} )".format),
    "lshift": ((lambda a1, a2: a1.typeName), "( {}<<{} )".format),
    "rshift": ((lambda a1, a2: a1.typeName), "( {}>>{} )".format),
    #
    "abs": (None, "std::abs( {} )".format),
    "sqrt": (_moTp_1EnsureFloat, "std::sqrt( {} )".format),
    "pow": ((lambda a1, a2: _moTp_1EnsureFloat(a1)), "std::pow( {}, {} )".format),
    "exp": (_moTp_1EnsureFloat, "std::exp( {} )".format),
    "log": (_moTp_1EnsureFloat, "std::log( {} )".format),
    "log10": (_moTp_1EnsureFloat, "std::log10( {} )".format),
    "sin": (_moTp_1EnsureFloat, "std::sin( {} )".format),
    "cos": (_moTp_1EnsureFloat, "std::cos( {} )".format),
    "tan": (_moTp_1EnsureFloat, "std::tan( {} )".format),
    "asin": (_moTp_1EnsureFloat, "std::asin( {} )".format),
    "acos": (_moTp_1EnsureFloat, "std::acos( {} )".format),
    "atan": (_moTp_1EnsureFloat, "std::atan( {} )".format),
    "max": (_moTp_minmax, "std::max( {}, {} )".format),
    "min": (_moTp_minmax, "std::min( {}, {} )".format),
    #
    "switch": ((lambda t, a, b: _moTp_minmax(a, b)), "( ( {} ) ? ( {} ) : ( {} ) )".format)
}


class MathOp(TupleOp):
    """ Mathematical function N->1, e.g. sin, abs, ( lambda x, y : x*y ) """
    @staticmethod
    def findResultType(op, args, hint=None):
        opStrat, _ = mathOpFuns_typeAndCppStr[op]
        fromStrat = None
        if isinstance(opStrat, str):  # operations that always return the same type, e.g. bool
            fromStrat = opStrat
        elif len(args) == 1 and opStrat is None:
            fromStrat = args[0].typeName  # unary -> same type by default
        elif opStrat:
            fromStrat = opStrat(*args)
        if fromStrat:
            return fromStrat
        if not hint:
            raise RuntimeError(f"No type could be inferred for MathOp {op!r}")
        return hint
    __slots__ = ("op", "args")

    def __init__(self, op, *args, **kwargs):
        self.op = op
        self.args = tuple(adaptArg(a, typeHint=floatType) for a in args)
        super().__init__(MathOp.findResultType(op, self.args, kwargs.pop("outType", None)))
        self.canDefine = kwargs.pop("canDefine", all(a.canDefine for a in self.args))
        assert len(kwargs) == 0

    def _clone(self, memo, select, selClones=None):
        argsCl = tuple(a.clone(memo=memo, select=select, selClones=selClones) for a in self.args)
        isSel = select(self)
        if isSel or any(id(aCl) != id(aOrig) for aCl, aOrig in zip(argsCl, self.args)):
            cln = self.__class__(self.op, *argsCl, outType=self.typeName, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in self.args:
                if select(arg):
                    yield arg
                yield from arg.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return "{}({}, {}, outType={!r})".format(
            self.__class__.__name__, self.op,
            ", ".join(repr(arg) for arg in self.args), self.typeName)

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.op, self.typeName, *self.args)

    def _eq(self, other):
        return self.typeName == other.typeName and self.op == other.op and self.args == other.args

    def get_cppStr(self, defCache=cppNoRedir):
        return mathOpFuns_typeAndCppStr[self.op][1](*(defCache(arg) for arg in self.args))


class GetItem(TupleOp):
    """ Get item from array (from function call or from array leaf) """
    __slots__ = ("arg", "_index")

    def __init__(self, arg, valueType, index, indexType=SizeType, canDefine=None):
        super().__init__(valueType)
        self.arg = adaptArg(arg)
        self._index = adaptArg(index, typeHint=indexType)
        self.canDefine = (canDefine if canDefine is not None
                          else (self.arg.canDefine and self._index.canDefine))

    def _clone(self, memo, select, selClones=None):
        argCl = self.arg.clone(memo=memo, select=select, selClones=selClones)
        idxCl = self._index.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if isSel or id(argCl) != id(self.arg) or id(self._index) != id(idxCl):
            cln = self.__class__(argCl, self.typeName, idxCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.arg, self._index):
                if select(arg):
                    yield arg
                yield from arg.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    @property
    def index(self):
        return self._index.result

    def _repr(self):
        return f"{self.__class__.__name__}({self.arg!r}, {self.typeName!r}, {self._index!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.arg, self.typeName, self._index)

    def _eq(self, other):
        return (self.arg == other.arg and self.typeName == other.typeName
                and self._index == other._index)

    def get_cppStr(self, defCache=cppNoRedir):
        return f"{defCache(self.arg)}[{defCache(self._index)}]"


class Construct(TupleOp):
    __slots__ = ("args",)

    def __init__(self, typeName, args, canDefine=None):
        super().__init__(typeName)
        self.args = tuple(adaptArg(a, typeHint=floatType) for a in args)
        self.canDefine = canDefine if canDefine is not None else all(a.canDefine for a in self.args)

    def _clone(self, memo, select, selClones=None):
        argsCl = tuple(a.clone(memo=memo, select=select, selClones=selClones) for a in self.args)
        isSel = select(self)
        if isSel or any(id(argCl) != id(arg) for argCl, arg in zip(argsCl, self.args)):
            cln = self.__class__(self.typeName, argsCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in self.args:
                if select(arg):
                    yield arg
                yield from arg.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return "{}({!r}, {})".format(
            self.__class__.__name__, self.typeName, ", ".join(repr(a) for a in self.args))

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.typeName, *self.args)

    def _eq(self, other):
        return self.typeName == other.typeName and self.args == other.args

    def get_cppStr(self, defCache=cppNoRedir):
        return "{0}{{{1}}}".format(self.typeName, ", ".join(defCache(a) for a in self.args))

    def getPyROOT(self, gbl):
        if len(self.args) == 1 and isinstance(self.args[0], InitList):
            ret = getattr(gbl, self.typeName)()
            for elm in self.args[0].elms:
                ret.push_back(elm.getPyROOT(gbl))
            return ret
        else:
            return getattr(gbl, self.typeName)(*(arg.getPyROOT(gbl) for arg in self.args))


def guessReturnType(mp):
    from .root import gbl
    oneDecl = None
    if hasattr(mp, "func_doc") and hasattr(mp, "func_name"):
        oneDecl = mp.func_doc.split("\n")[0]  # overloads should have the same return type
    elif ((hasattr(gbl, "TemplateProxy") and isinstance(mp, gbl.TemplateProxy))
          or (type(mp).__name__ == "TemplateProxy")):
        oneDecl = mp.__doc__.split("\n")[0]
    if oneDecl:
        toks = list(oneDecl.split())
        # left and right strip const * and &
        while toks[-1].rstrip("&") in ("", "const", "static"):
            toks = toks[:-1]
        while toks[0].rstrip("&") in ("", "const", "static"):
            toks = toks[1:]
        while any(tok.endswith("unsigned") for tok in toks):
            iU = next(i for i, tok in enumerate(toks) if tok.endswith("unsigned"))
            toks[iU] = " ".join((toks[iU], toks[iU + 1]))
            del toks[iU + 1]
        if len(toks) == 2:
            return toks[0].rstrip("&")
        else:
            nOpen = 0
            i = 0
            while i < len(toks) and (i == 0 or nOpen != 0):
                nOpen += (toks[i].count("<") - toks[i].count(">"))
                i += 1
            return " ".join(toks[:i]).rstrip("&")
    else:
        return "Float_t"


class CallMethod(TupleOp):
    """
    Call a method
    """
    __slots__ = ("name", "args")

    def __init__(self, name, args, returnType=None, getFromRoot=True, canDefine=None):
        super().__init__(returnType if returnType
                         else CallMethod._initReturnType(name, getFromRoot=getFromRoot))
        self.name = name  # NOTE can only be a hardcoded string this way
        self.args = tuple(adaptArg(arg) for arg in args)
        self.canDefine = (canDefine if canDefine is not None
                          else all(a.canDefine for a in self.args))

    @staticmethod
    def _initReturnType(name, getFromRoot=True):
        mp = None
        if getFromRoot:
            try:
                from .root import gbl
                if "::" in name:
                    res = gbl
                    for tok in name.split("::"):
                        res = getattr(res, tok)
                    if res != gbl:
                        mp = res
                else:
                    mp = getattr(gbl, name)
            except Exception as ex:
                logger.error(f"Exception in getting method pointer {name}: {ex}")
        return guessReturnType(mp)

    def _clone(self, memo, select, selClones=None):
        argsCl = tuple(a.clone(memo=memo, select=select, selClones=selClones) for a in self.args)
        isSel = select(self)
        if isSel or any(id(argCl) != id(arg) for argCl, arg in zip(argsCl, self.args)):
            cln = self.__class__(self.name, argsCl, returnType=self.typeName, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.add(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in self.args:
                if select(arg):
                    yield arg
                yield from arg.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return "{}({!r}, ({}), returnType={!r})".format(
            self.__class__.__name__, self.name,
            ", ".join(repr(arg) for arg in self.args), self.typeName)

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.name, self.typeName, *self.args)

    def _eq(self, other):
        return self.name == other.name and self.typeName == other.typeName and self.args == other.args

    # backends
    def get_cppStr(self, defCache=cppNoRedir):
        return "{}({})".format(self.name, ", ".join(defCache(arg) for arg in self.args))


class CallMemberMethod(TupleOp):
    """ Call a member method """
    __slots__ = ("this", "name", "args", "byPointer")

    def __init__(self, this, name, args, byPointer=False, returnType=None, canDefine=None):
        self.this = adaptArg(this)
        self.name = name  # NOTE can only be a hardcoded string this way
        self.args = tuple(adaptArg(arg) for arg in args)
        self.byPointer = byPointer
        super().__init__(returnType if returnType
                         else guessReturnType(getattr(this._typ, self.name)))
        self.canDefine = (canDefine if canDefine is not None
                          else self.this.canDefine and all(a.canDefine for a in self.args))

    def _clone(self, memo, select, selClones=None):
        thisCl = self.this.clone(memo=memo, select=select, selClones=selClones)
        argsCl = tuple(a.clone(memo=memo, select=select, selClones=selClones)
                       for a in self.args)
        isSel = select(self)
        if (isSel or id(thisCl) != id(self.this)
                or any(id(argCl) != id(arg) for argCl, arg in zip(argsCl, self.args))):
            cln = self.__class__(
                thisCl, self.name, argsCl, byPointer=self.byPointer,
                returnType=self.typeName, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in chain((self.this,), self.args):
                if select(arg):
                    yield arg
                yield from arg.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return "{}({!r}, {!r}, ({}), byPointer={!r}, returnType={!r})".format(
            self.__class__.__name__, self.this, self.name,
            ", ".join(repr(arg) for arg in self.args), self.byPointer, self.typeName)

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.this, self.name,
                   repr(self.byPointer), self.typeName, *self.args)

    def _eq(self, other):
        return (self.this == other.this and self.name == other.name
                and self.byPointer == other.byPointer
                and self.typeName == other.typeName and self.args == other.args)

    def get_cppStr(self, defCache=cppNoRedir):
        if self.name == "__call__":
            call = ""
        elif self.byPointer:
            call = f"->{self.name}"
        else:
            call = f".{self.name}"
        return "{}{}({})".format(
            defCache(self.this), call, ", ".join(defCache(arg) for arg in self.args))


class GetDataMember(TupleOp):
    """ Get a data member """
    __slots__ = ("this", "name", "byPointer")

    def __init__(self, this, name, byPointer=False, returnType=None, canDefine=None):
        self.this = adaptArg(this)
        self.name = name  # NOTE can only be a hardcoded string this way
        self.byPointer = byPointer
        super().__init__(returnType if returnType else GetDataMember._initType(this, name))
        self.canDefine = canDefine if canDefine is not None else self.this.canDefine

    @staticmethod
    def _initType(this, name):
        try:
            protoTp = this._typ
            proto = protoTp()  # should *in principle* work for most ROOT objects
            att = getattr(proto, name)
            tpNm = type(att).__name__
            if protoTp.__name__.startswith("pair<") and name in ("first", "second"):
                tpNms = tuple(tok.strip() for tok in protoTp.__name__[5:-1].split(","))
                return (tpNms[0] if name == "first" else tpNms[1])
            return tpNm
        except Exception as e:
            logger.error(f"Problem getting type of data member {name} of {this!r}", e)
            return "void"

    def _clone(self, memo, select, selClones=None):
        thisCl = self.this.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if isSel or id(thisCl) != id(self.this):
            cln = self.__class__(
                thisCl, self.name, byPointer=self.byPointer,
                returnType=self.typeName, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            if select(self.this):
                yield self.this
            yield from self.this.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return f"{self.__class__.__name__}({self.this!r}, {self.name!r}, {self.byPointer!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.this, self.name, repr(self.byPointer))

    def _eq(self, other):
        return (self.this == other.this and self.name == other.name
                and self.byPointer == other.byPointer)

    def get_cppStr(self, defCache=cppNoRedir):
        return "{}{}{}".format(defCache(self.this), ("->" if self.byPointer else "."), self.name)


class ExtVar(TupleOp):
    """ Externally-defined variable (used by name) """
    __slots__ = ("name",)

    def __init__(self, typeName, name):
        super().__init__(typeName)
        self.name = name

    def _clone(self, memo, select, selClones=None):
        if select(self):
            cln = self.__class__(self.typeName, self.name)
            if selClones is not None:
                selClones.append(cln)
            return cln

    def _repr(self):
        return f"{self.__class__.__name__}({self.typeName!r}, {self.name!r})"

    def _hash(self, fun=op_hash):
        return fun(self._repr())

    def _eq(self, other):
        return self.typeName == other.typeName and self.name == other.name

    def get_cppStr(self, defCache=None):
        return self.name


class DefinedSymbol(TupleOp):
    """ Defined variable (used by name), first use will trigger definition """
    __slots__ = ("definition", "_nameHint")

    def __init__(self, typeName, definition, nameHint=None):
        super().__init__(typeName)
        self.definition = definition
        self._nameHint = nameHint

    def _clone(self, memo, select, selClones=None):
        if select(self):
            cln = self.__class__(self.typeName, self.definition, nameHint=self._nameHint)
            if selClones is not None:
                selClones.append(cln)
            return cln

    def _repr(self):
        return (f"{self.__class__.__name__}({self.typeName!r}, {self.definition!r}, "
                f"nameHint={self._nameHint!r})")

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.typeName,
                   self.definition, (self._nameHint or 'None'))

    def _eq(self, other):
        return (self.typeName == other.typeName and self.definition == other.definition
                and self._nameHint == other._nameHint)

    def get_cppStr(self, defCache=cppNoRedir):
        return defCache.symbol(self.definition, nameHint=self._nameHint)


class InitList(TupleOp):
    """ Initializer list """
    __slots__ = ("elms",)

    def __init__(self, typeName, elms, elmType=None, canDefine=None):
        super().__init__(typeName)
        self.typeName = typeName
        self.elms = tuple(adaptArg(e, typeHint=elmType) for e in elms)
        self.canDefine = (canDefine if canDefine is not None
                          else all(elm.canDefine for elm in self.elms))

    def _clone(self, memo, select, selClones=None):
        elmsCl = tuple(elm.clone(memo=memo, select=select, selClones=selClones)
                       for elm in self.elms)
        isSel = select(self)
        if isSel or any(id(elmCl) != id(elm) for elmCl, elm in zip(elmsCl, self.elms)):
            cln = self.__class__(self.typeName, elmsCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for elm in self.elms:
                if select(elm):
                    yield elm
                yield from elm.deps(defCache=defCache, select=select, includeLocal=includeLocal)

    def _repr(self):
        return "{}<{}>({})".format(
            self.__class__.__name__, self.typeName,
            ", ".join(repr(elm) for elm in self.elms))

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.typeName, *self.elms)

    def _eq(self, other):
        return self.typeName == other.typeName and self.elms == other.elms

    def get_cppStr(self, defCache=cppNoRedir):
        return "{{ {0} }}".format(", ".join(defCache(elm) for elm in self.elms))


class LocalVariablePlaceholder(TupleOp):
    """
    Placeholder type for a local variable connected to an index
        (first step in a specific-to-general strategy)
    """
    __slots__ = ("_parent", "i")

    def __init__(self, typeName, parent=None, i=None):
        super().__init__(typeName)
        self._parent = parent
        self.i = i  # FIXME this one is set **late** - watch out with what we call
        self.canDefine = False

    def _clone(self, memo, select, selClones=None):
        if select(self):
            raise RuntimeError("LocalVariablePlaceholder should not be cloned")
            return self.__class__(self.typeName, parent=self._parent, i=self.i)

    @property
    def name(self):
        if self.i is None:
            raise RuntimeError("Using LocalVariablePlaceholder before giving it an index")
        return f"i{self.i:d}"

    def get_cppStr(self, defCache=None):
        return self.name

    def _repr(self):
        return f"{self.__class__.__name__}({self.typeName!r}, i={self.i!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.typeName, str(self.i))

    def _eq(self, other):
        # fine for the LocalVariablePlaceholder to be equal from different sub-expressions because
        # it will never be defined and if equal to an expression that can be defined, it's fine
        return self.typeName == other.typeName and self.i is not None and self.i == other.i


def collectNodes(expr, defCache=cppNoRedir, select=(lambda nd: True), includeLocal=True):
    # simple helper
    if select(expr):
        yield expr
    yield from expr.deps(defCache=defCache, select=select, includeLocal=includeLocal)


def _triggerDefinitions(self, defCache=cppNoRedir):
    for dep in self.deps(defCache=defCache, select=lambda op: defCache.shouldDefine(op)):
        cn = defCache(dep)
        if not cn:
            logger.warning(f"Probably a problem in triggering definition for {dep}")


def _collectDeps(self, defCache=cppNoRedir):
    return {dep for dep in self.deps(defCache=defCache, select=(
        lambda op: (isinstance(op, GetColumn) or isinstance(op, GetArrayColumn)
                    or isinstance(op, LocalVariablePlaceholder) or isinstance(op, Parameter)
                    or defCache.shouldDefine(op) or (defCache.getColName(op) is not None))))}


class StopAt:
    def __init__(self, stopPred):
        self.stopPred = stopPred

    def __call__(self, arg):
        raise NotImplementedError("This is only to control the loops")

    def symbol(self, decl):
        raise NotImplementedError(
            f"Asked to add defined symbol for {decl}, but that's not supported; "
            "this is only to control the loop")

    def stop(self, op):
        return self.stopPred(op)

    def _getColNme(self, op):
        return None


class VisitOnceAndStopAt(StopAt):
    def __init__(self, stopPred=None):
        self.visitedIds = set()
        super().__init__(stopPred)

    def stop(self, op):
        opId = id(op)
        seenBefore = (opId in self.visitedIds)
        if not seenBefore:
            self.visitedIds.add(opId)
        return seenBefore or (self.stopPred is not None and self.stopPred(op))


# replacement for cases that don't have the full definition cache (e.g. during construction)
stopAtDefinedOrCan = StopAt(lambda op: op.canDefine or isinstance(op, DefineOnFirstUse))


def _canDefine(expr, local):
    return not any(ind is not None for ind in collectNodes(
        expr, defCache=stopAtDefinedOrCan, includeLocal=False,
        select=(lambda nd: isinstance(nd, LocalVariablePlaceholder) and nd not in local)))


def _isRangeAndHasMxi(op):
    return isinstance(op, RangeOp) and op._mxi is not None


def _maxLVIdx(expr):
    return max(chain([-1], ((nd._mxi if nd._mxi is not None else -1) for nd in collectNodes(
        expr, defCache=VisitOnceAndStopAt(_isRangeAndHasMxi), select=_isRangeAndHasMxi))))


def _convertFunArgs(deps, defCache=cppNoRedir):
    from . import treeproxies as _tp
    capDeclCall = []
    for ld in deps:
        if isinstance(ld, GetArrayColumn):
            capDeclCall.append((f"&{ld.name}",
                                f"const ROOT::VecOps::RVec<{ld.valueType}>& {ld.name}",
                                ld.name, ld.name))
        elif isinstance(ld, GetColumn):
            if any(ld.typeName in basicTp for basicTp in (
                    _tp._boolTypes, _tp._integerTypes, _tp._floatTypes)):  # by value
                capDeclCall.append((ld.name, f"{ld.typeName} {ld.name}", ld.name, ld.name))
            else:  # by const-ref
                capDeclCall.append((f"&{ld.name}",
                                    f"const {ld.typeName}& {ld.name}",
                                    ld.name, ld.name))
        elif isinstance(ld, LocalVariablePlaceholder):
            if not ld.name:
                print(f"ERROR: no name for local {ld}")
            capDeclCall.append((ld.name, f"{ld.typeName} {ld.name}", ld.name, ld.name))
        elif defCache.shouldDefine(ld) or (defCache.getColName(ld) is not None):
            nm = defCache.getColName(ld)
            if not nm:
                print(f"ERROR: no column name for {ld}")
            if not any(f"&{nm}" == icap[0] for icap in capDeclCall):
                capDeclCall.append((f"&{nm}", f"const {ld.result._typeName}& {nm}", nm, nm))
            else:
                print(f"WARNING: dependency {nm} is there twice")
        elif isinstance(ld, Parameter):
            capDeclCall.append((ld.name, f"{ld.typeName} {ld.name}", ld.name, str(ld.value)))
        else:
            raise AssertionError(f"Dependency with unknown type: {ld}")
    if capDeclCall:
        # sort by declaration (alphabetic for the same type)
        return zip(*sorted(capDeclCall, key=(lambda elm: elm[1])))
    else:
        return [], [], [], []


class RangeOp(TupleOp):
    __slots__ = ("rng",)

    def __init__(self, rng, typeName):
        self.rng = rng
        super().__init__(typeName)

    @property
    def _mxi(self):
        return None


class Select(RangeOp):
    """ Define a selection on a range """
    __slots__ = ("predExpr", "_i")

    def __init__(self, rng, predExpr, idx, canDefine=None):
        super().__init__(rng, f"ROOT::VecOps::RVec<{SizeType}>")
        self.predExpr = predExpr
        self._i = idx
        self.canDefine = (
            canDefine if canDefine is not None
            else self.rng.canDefine and _canDefine(self.predExpr, (self._i,)))

    @property
    def _mxi(self):
        return self._i.i

    def _clone(self, memo, select, selClones=None):
        rngCl = self.rng.clone(memo=memo, select=select, selClones=selClones)
        predCl = self.predExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = self._i.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if (isSel or id(rngCl) != id(self.rng) or id(predCl) != id(self.predExpr)
                or id(iCl) != id(self._i)):
            cln = self.__class__(rngCl, predCl, iCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    @staticmethod
    def fromRngFun(rng, pred):
        """ Factory method from a range and predicate (callable) """
        idx = LocalVariablePlaceholder(SizeType)
        predExpr = adaptArg(pred(rng._getItem(idx.result)))
        idx.i = _maxLVIdx(predExpr) + 1
        res = Select(adaptArg(rng.idxs), predExpr, idx)
        idx._parent = res
        from .treeproxies import SelectionProxy
        return SelectionProxy(rng._base, res, valueType=rng.valueType)

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.rng, self.predExpr):
                if select(arg) and (includeLocal or arg != self._i):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp != self._i:
                        yield dp

    @property
    def result(self):
        from .treeproxies import VectorProxy
        return VectorProxy(self, itemType=SizeType)

    def _repr(self):
        return f"{self.__class__.__name__}({self.rng!r}, {self.predExpr!r}, {self._i!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.rng, self.predExpr, self._i)

    def _eq(self, other):
        return self.rng == other.rng and self.predExpr == other.predExpr and self._i == other._i

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return "rdfhelpers::select({idxs},\n    [{captures}] ( {i} ) {{ return {predExpr}; }})".format(
            idxs=defCache(self.rng),
            captures=",".join(captures),
            i=f"{self._i.typeName} {self._i.name}",
            predExpr=defCache(self.predExpr)
        )


class Sort(RangeOp):
    """ Sort a range (ascendingly) by the value of a function on each element """
    __slots__ = ("funExpr", "_i")

    def __init__(self, rng, funExpr, idx, canDefine=None):
        super().__init__(rng, f"ROOT::VecOps::RVec<{SizeType}>")
        self.funExpr = funExpr
        self._i = idx
        self.canDefine = (
            canDefine if canDefine is not None
            else self.rng.canDefine and _canDefine(self.funExpr, (self._i,)))

    @property
    def _mxi(self):
        return self._i.i

    def _clone(self, memo, select, selClones=None):
        rngCl = self.rng.clone(memo=memo, select=select, selClones=selClones)
        funCl = self.funExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = self._i.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if (isSel or id(rngCl) != id(self.rng) or id(funCl) != id(self.funExpr)
                or id(iCl) != id(self._i)):
            cln = self.__class__(rngCl, funCl, iCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    @staticmethod
    def fromRngFun(rng, fun):
        idx = LocalVariablePlaceholder(SizeType)
        funExpr = adaptArg(fun(rng._getItem(idx.result)))
        idx.i = _maxLVIdx(funExpr) + 1
        res = Sort(adaptArg(rng.idxs), funExpr, idx)
        idx._parent = res
        from .treeproxies import SelectionProxy
        return SelectionProxy(rng._base, res, valueType=rng.valueType)

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.rng, self.funExpr):
                if select(arg) and (includeLocal or arg != self._i):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp != self._i:
                        yield dp

    @property
    def result(self):
        from .treeproxies import VectorProxy
        return VectorProxy(self, itemType=SizeType)

    def _repr(self):
        return f"{self.__class__.__name__}({self.rng!r}, {self.funExpr!r}, {self._i!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.rng, self.funExpr, self._i)

    def _eq(self, other):
        return self.rng == other.rng and self.funExpr == other.funExpr and self._i == other._i

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return "rdfhelpers::sort({idxs},\n    [{captures}] ( {i} ) {{ return {funExpr}; }})".format(
            idxs=defCache(self.rng),
            captures=",".join(captures),
            i=f"{self._i.typeName} {self._i.name}",
            funExpr=defCache(self.funExpr)
        )


class Map(RangeOp):
    """ Create a list of derived values for a collection (mostly useful for storing on skims) """
    __slots__ = ("funExpr", "_i", "mappedType")

    def __init__(self, rng, funExpr, idx, mappedType, canDefine=None):
        # self.rng are the input indices
        super().__init__(rng, f"ROOT::VecOps::RVec<{mappedType}>")
        self.funExpr = funExpr
        self._i = idx
        self.mappedType = mappedType
        self.canDefine = (
            canDefine if canDefine is not None
            else self.rng.canDefine and _canDefine(self.funExpr, (self._i,)))

    @property
    def _mxi(self):
        return self._i.i

    def _clone(self, memo, select, selClones=None):
        rngCl = self.rng.clone(memo=memo, select=select, selClones=selClones)
        funCl = self.funExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = self._i.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if (isSel or id(rngCl) != id(self.rng) or id(funCl) != id(self.funExpr)
                or id(iCl) != id(self._i)):
            cln = self.__class__(rngCl, funCl, iCl, self.mappedType, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    @staticmethod
    def fromRngFun(rng, fun, typeName=None):
        idx = LocalVariablePlaceholder(SizeType)
        val = fun(rng._getItem(idx.result))
        funExpr = adaptArg(val)
        idx.i = _maxLVIdx(funExpr) + 1
        res = Map(adaptArg(rng.idxs), funExpr, idx,
                  (typeName if typeName is not None else val._typeName))
        idx._parent = res
        return res.result

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.rng, self.funExpr):
                if select(arg) and (includeLocal or arg != self._i):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp != self._i:
                        yield dp

    @property
    def result(self):
        from .treeproxies import VectorProxy
        return VectorProxy(self, itemType=self.mappedType)

    def _repr(self):
        return f"{self.__class__.__name__}({self.rng!r}, {self.funExpr!r}, {self._i!r}, {self.mappedType!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.rng, self.funExpr, self._i, self.mappedType)

    def _eq(self, other):
        return (self.rng == other.rng and self.funExpr == other.funExpr
                and self._i == other._i and self.mappedType == other.mappedType)

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return "rdfhelpers::map<{valueType}>({idxs},\n    [{capt}] ( {i} ) {{ return {funExpr}; }})".format(
            valueType=self.mappedType,
            idxs=defCache(self.rng),
            capt=",".join(captures),
            i=f"{self._i.typeName} {self._i.name}",
            funExpr=defCache(self.funExpr)
        )


class Next(RangeOp):
    """ Define a search (first matching item, for a version that processes the whole range see Reduce) """
    __slots__ = ("predExpr", "_i")

    def __init__(self, rng, predExpr, idx, canDefine=None):
        # self.rng are the input indices
        super().__init__(rng, SizeType)
        self.predExpr = predExpr
        self._i = idx
        self.canDefine = (
            canDefine if canDefine is not None
            else self.rng.canDefine and _canDefine(self.predExpr, (self._i,)))

    @property
    def _mxi(self):
        return self._i.i

    def _clone(self, memo, select, selClones=None):
        rngCl = self.rng.clone(memo=memo, select=select, selClones=selClones)
        predCl = self.predExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = self._i.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if (isSel or id(rngCl) != id(self.rng) or id(predCl) != id(self.predExpr)
                or id(iCl) != id(self._i)):
            cln = self.__class__(rngCl, predCl, iCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    @staticmethod
    def fromRngFun(rng, pred):  # FIXME you are here
        idx = LocalVariablePlaceholder(SizeType)
        predExpr = adaptArg(pred(rng._getItem(idx.result)))
        idx.i = _maxLVIdx(predExpr) + 1
        res = Next(adaptArg(rng.idxs), predExpr, idx)
        idx._parent = res
        return rng._getItem(res)

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.rng, self.predExpr):
                if select(arg) and (includeLocal or arg != self._i):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp != self._i:
                        yield dp

    def _repr(self):
        return f"{self.__class__.__name__}({self.rng!r}, {self.predExpr!r}, {self._i!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.rng, self.predExpr, self._i)

    def _eq(self, other):
        return self.rng == other.rng and self.predExpr == other.predExpr and self._i == other._i

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return "rdfhelpers::next({idxs},\n     [{captures}] ( {i} ) {{ return {predexpr}; }}, -1)".format(
            idxs=defCache(self.rng),
            captures=",".join(captures),
            i=f"{self._i.typeName} {self._i.name}",
            predexpr=defCache(self.predExpr),
        )


class Reduce(RangeOp):
    """ Reduce a range to a value (could be a transformation, index...) """
    __slots__ = ("start", "accuExpr", "_i", "_prevRes")

    def __init__(self, rng, resultType, start, accuExpr, idx, prevRes, canDefine=None):
        # self.rng are the input indices
        super().__init__(rng, resultType)
        self.start = start
        self.accuExpr = accuExpr
        self._i = idx
        self._prevRes = prevRes
        self.canDefine = (
            canDefine if canDefine is not None
            else self.rng.canDefine and start.canDefine
            and _canDefine(accuExpr, (self._i, self._prevRes)))

    @property
    def _mxi(self):
        return self._prevRes.i

    def _clone(self, memo, select, selClones=None):
        rngCl = self.rng.clone(memo=memo, select=select, selClones=selClones)
        startCl = self.start.clone(memo=memo, select=select, selClones=selClones)
        accuCl = self.accuExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = self._i.clone(memo=memo, select=select, selClones=selClones)
        prevCl = self._prevRes.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if (isSel or id(rngCl) != id(self.rng) or id(startCl) != id(self.start)
                or id(accuCl) != id(self.accuExpr) or id(iCl) != id(self._i)
                or id(prevCl) != id(self._prevRes)):
            cln = self.__class__(rngCl, self.typeName, startCl, accuCl,
                                 iCl, prevCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(self)
            return cln

    @staticmethod
    def fromRngFun(rng, start, accuFun):
        resultType = start._typeName
        idx = LocalVariablePlaceholder(SizeType)
        prevRes = LocalVariablePlaceholder(resultType, i=-1)
        accuExpr = adaptArg(accuFun(prevRes.result, rng._getItem(idx.result)))
        maxLVIdx = _maxLVIdx(accuExpr)
        idx.i = maxLVIdx + 1
        prevRes.i = maxLVIdx + 2

        res = Reduce(adaptArg(rng.idxs), resultType, adaptArg(start), accuExpr, idx, prevRes)
        idx._parent = res
        prevRes._parent = res
        return res.result

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in (self.rng, self.start, self.accuExpr):
                if select(arg) and (includeLocal or arg not in (self._i, self._prevRes)):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp not in (self._i, self._prevRes):
                        yield dp

    def _repr(self):
        return (f"{self.__class__.__name__}({self.rng!r}, {self.typeName!r}, "
                f"{self.start!r}, {self.accuExpr!r}, {self._i!r}, {self._prevRes!r})")

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.rng, self.typeName,
                   self.start, self.accuExpr, self._i, self._prevRes)

    def _eq(self, other):
        return (self.rng == other.rng and self.typeName == other.typeName
                and self.start == other.start and self.accuExpr == other.accuExpr
                and self._i == other._i and self._prevRes == other._prevRes)

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return ("rdfhelpers::reduce({idxs}, {start},\n"
                "     [{captures}] ( {prevRes}, {i} ) {{ return {accuexpr}; }})").format(
            idxs=defCache(self.rng),
            start=defCache(self.start),
            captures=",".join(captures),
            prevRes=f"{self._prevRes.typeName} {self._prevRes.name}",
            i=f"{self._i.typeName} {self._i.name}",
            accuexpr=defCache(self.accuExpr)
        )


class Combine(RangeOp):
    __slots__ = ("candPredExpr", "_i")

    def __init__(self, ranges, candPredExpr, idx, canDefine=None):
        # self.rng is [ input index lists ]
        super().__init__(ranges, f"ROOT::VecOps::RVec<rdfhelpers::Combination<{len(ranges):d}>>")
        self.candPredExpr = candPredExpr
        self._i = idx
        self.canDefine = (
            canDefine if canDefine is not None
            else all(rng.canDefine for rng in self.rng) and _canDefine(candPredExpr, self._i))

    @property
    def _mxi(self):
        return self._i[-1].i

    @property
    def n(self):
        return len(self.rng)

    def _clone(self, memo, select, selClones=None):
        rngCl = tuple(rng.clone(memo=memo, select=select, selClones=selClones) for rng in self.rng)
        predCl = self.candPredExpr.clone(memo=memo, select=select, selClones=selClones)
        iCl = tuple(i.clone(memo=memo, select=select, selClones=selClones) for i in self._i)
        isSel = select(self)
        if (isSel or any(id(rCl) != id(rng) for rCl, rng in zip(rngCl, self.rng))
                or id(predCl) != id(self.candPredExpr)
                or any(id(iiCl) != id(iOr) for iiCl, iOr in zip(iCl, self._i))):
            cln = self.__class__(rngCl, predCl, iCl, canDefine=self.canDefine)
            if selClones is not None and isSel:
                selClones.append(isSel)
            return cln

    @staticmethod
    def fromRngFun(num, ranges, candPredFun, samePred=None):
        ranges = ranges if len(ranges) > 1 else list(repeat(ranges[0], num))
        idx = tuple(LocalVariablePlaceholder(SizeType, i=-1 - i) for i in range(num))
        candPred = candPredFun(*(rng._getItem(iidx.result) for rng, iidx in zip(ranges, idx)))
        candPredExpr = adaptArg(candPred)
        if samePred:
            from . import treefunctions as op
            areDiff = op.AND(*(
                samePred(ra._getItem(ia.result), rb._getItem(ib.result))
                for ((ia, ra), (ib, rb)) in combinations(zip(idx, ranges), 2)
                if ra._base == rb._base))
            if len(areDiff.op.args) > 0:
                candPredExpr = adaptArg(op.AND(areDiff, candPred))
        maxLVIdx = _maxLVIdx(candPredExpr)
        for i, ilvp in enumerate(idx):
            ilvp.i = maxLVIdx + 1 + i
        res = Combine(tuple(adaptArg(rng.idxs) for rng in ranges), candPredExpr, idx)
        for ilvp in idx:
            ilvp._parent = res
        from .treeproxies import CombinationListProxy
        return CombinationListProxy(ranges, res)

    def deps(self, defCache=cppNoRedir, select=(lambda x: True), includeLocal=False):
        if not defCache.stop(self):
            for arg in chain(self.rng, [self.candPredExpr]):
                if select(arg) and (includeLocal or arg not in self._i):
                    yield arg
                for dp in arg.deps(defCache=defCache, select=select, includeLocal=includeLocal):
                    if includeLocal or dp not in self._i:
                        yield dp

    def _repr(self):
        return f"{self.__class__.__name__}({self.rng!r}, {self.candPredExpr!r}, {self._i!r})"

    def _hash(self, fun=op_hash):
        return fun(*chain([self.__class__.__name__], self.rng, [self.candPredExpr], self._i))

    def _eq(self, other):
        return (equal_elements(self.rng, other.rng) and self.candPredExpr == other.candPredExpr
                and self._i == other._i)

    def get_cppStr(self, defCache=cppNoRedir):
        _triggerDefinitions(self, defCache=defCache)
        captures, _, _, _ = _convertFunArgs(_collectDeps(self, defCache=defCache), defCache=defCache)
        return (
            "rdfhelpers::combine(\n"
            "     [{captures}] ( {predIdxArgs} ) {{ return {predExpr}; }},\n"
            "     {ranges})").format(
                captures=",".join(captures),
                predIdxArgs=", ".join(f"{i.typeName} {i.name}" for i in self._i),
                predExpr=defCache(self.candPredExpr),
                ranges=", ".join(defCache(rng) for rng in self.rng)
        )


# FIXME to be implemented
class PseudoRandom(TupleOp):
    """ Pseudorandom number (integer or float) within range """
    def __init__(self, xMin, xMax, seed, isIntegral=False):
        super().__init__()
        self.xMin = xMin
        self.xMax = xMax
        self.seed = seed
        self.isIntegral = isIntegral

    @property
    def resultType(self):
        return "Int_" if self.isIntegral else "Float_t"
    # deps from xMin, xMax and seed
    # seed can be event-based or object-based, depending?
    # TODO implement C++ side as well


class OpWithSyst(ForwardingOp):
    """
    Interface and base class for nodes that can change the systematic variation
        of something they wrap
    """
    def __init__(self, wrapped, variations=None):
        super().__init__(wrapped)
        self.variations = variations

    def _clone(self, memo, select, selClones=None):
        clWr = self.wrapped.clone(memo=memo, select=select, selClones=selClones)
        isSel = select(self)
        if isSel or id(clWr) != id(self.wrapped):
            cln = self.__class__(clWr, variations=self.variations)
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def changeVariation(self, newVariation):
        """ Assumed to be called on a fresh copy - *will* change the underlying value
        """
        if self._cache:  # validate this assumption
            raise RuntimeError("Cannot change variation of an expression that is already frozen")
        if newVariation not in self.variations:
            raise ValueError(f"Invalid variation: {newVariation}")
        self._changeVariation(newVariation)

    def _changeVariation(self, newVariation):
        """
        changeVariation specifics (after validating newVariation, and changability) -
            to be implemented by concrete classes
        """
        pass

    def _repr(self):
        return "{}({!r}, {!r}, {!r})".format(self.__class__.__name__, self.wrapped, self.variations)

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.wrapped, str(self.variations))

    def _eq(self, other):
        return super()._eq(other) and self.variations == other.variations


class SystAltOp(OpWithSyst):
    """ Change the wrapped operation (from a map) """
    def __init__(self, wrapped, varMap, valid=None):
        super().__init__(wrapped)
        self.variations = valid if valid is not None else tuple(varMap.keys())
        self.varMap = varMap

    def _clone(self, memo, select, selClones=None):
        clWr = self.wrapped.clone(memo=memo, select=select, selClones=selClones)
        varCl = {nm: vop.clone(memo=memo, select=select, selClones=selClones)
                 for nm, vop in self.varMap.items()}
        isSel = select(self)
        if (isSel or id(clWr) != id(self.wrapped)
                or any(id(vCl) != id(self.varMap[ky]) for ky, vCl in varCl.items())):
            cln = self.__class__(clWr, varCl, valid=tuple(self.variations))
            if selClones is not None and isSel:
                selClones.append(cln)
            return cln

    def _repr(self):
        return f"{self.__class__.__name__}({self.wrapped!r}, {self.variations!r}, {self.varMap!r})"

    def _hash(self, fun=op_hash):
        return fun(self.__class__.__name__, self.wrapped, str(self.variations),
                   *chain.from_iterable(self.varMap.items()))

    def _eq(self, other):
        return super()._eq(other) and self.variations == other.variations and self.varMap == other.varMap

    def _changeVariation(self, newVariation):
        if newVariation in self.varMap:
            self.wrapped = self.varMap[newVariation]
            self.typeName = self.wrapped.typeName


class DefineOnFirstUse(ForwardingOp):
    """ Node marked for definition as a column on first use """
    __slots__ = tuple()

    def __init__(self, wrapped, canDefine=None):
        super().__init__(wrapped, canDefine=canDefine)


def collectSystVars(exprs):
    # { varName : { expr : [ nodes to change ] } }
    systVars = defaultdict(dict)
    for expr in exprs:
        toChangeForThis = defaultdict(list)
        for nd in set(collectNodes(expr,
                                   select=(lambda nd_: isinstance(nd_, OpWithSyst) and nd_.variations),
                                   defCache=VisitOnceAndStopAt())):
            for vari in nd.variations:
                toChangeForThis[vari].append(nd)
        for vari, variNodes in toChangeForThis.items():
            systVars[vari][expr] = variNodes
    return systVars



"""
The :py:mod:`bamboo.scalefactors` module contains helper methods
for configuring scale factors, fake rates etc.

The basic configuration parameter is the JSON file path for a set of scalefactors.
There two basic types are

- lepton scale factors (dependent on a number of object variables, e.g. pt and eta),
- jet (b-tagging) scale factors (grouped set for different flavours, for convenience)

Different values (depending on the data-taking period)
can be taken into account by weighting or by randomly sampling.
"""
__all__ = ("get_scalefactor", "lumiPerPeriod_default", "binningVariables_nano", "BtagSF", "get_correction",
           "get_bTagSF_fixWP", "makeBtagWeightMeth1a", "get_bTagSF_itFit", "makeBtagWeightItFit")

import logging
from functools import partial
from itertools import chain

from . import treefunctions as op

logger = logging.getLogger(__name__)


#: Integrated luminosity (in 1/pb) per data taking period
lumiPerPeriod_default = {
    # 2016 - using approved normtag + 07Aug2017 re-reco golden JSON
    # ReReco/Final/Cert_271036-284044_13TeV_ReReco_07Aug2017_Collisions16_JSON.txt
    "Run2016B": 5750.491,
    "Run2016C": 2572.903,
    "Run2016D": 4242.292,
    "Run2016E": 4025.228,
    "Run2016F": 3104.509,
    "Run2016G": 7575.824,
    "Run2016H": 8650.628,
    # hww muon periods
    "Run271036to275783": 6274.191,
    "Run275784to276500": 3426.131,
    "Run276501to276811": 3191.207,

    # 2017 - using approved normtag + re-reco golden JSON
    # ReReco/Cert_294927-306462_13TeV_EOY2017ReReco_Collisions17_JSON_v1.txt
    "Run2017B": 4793.970,
    "Run2017C": 9632.746,
    "Run2017D": 4247.793,
    "Run2017E": 9314.581,
    "Run2017F": 13539.905,

    # 2018 - using approved normtag + re-reco golden JSON
    # ReReco/Cert_314472-325175_13TeV_17SeptEarlyReReco2018ABC_PromptEraD_Collisions18_JSON.txt
    "Run2018A": 14027.614,
    "Run2018B": 7066.552,
    "Run2018C": 6898.817,
    "Run2018D": 31747.582,
    # before/after muon HLT update (during 2018A)
    "Run315264to316360": 8928.970,
    "Run316361to325175": 50789.746
}


# TODO maybe move this elsewhere
binningVariables_nano = {
    "Eta": lambda obj: obj.eta,
    "AbsEta": lambda obj: op.abs(obj.eta),
    "ClusEta": lambda el: el.eta + el.deltaEtaSC,
    "AbsClusEta": lambda el: op.abs(el.eta + el.deltaEtaSC),
    "Pt": lambda obj: obj.pt
}


def getBinningVarNames(jsonpath):
    import json
    with open(jsonpath) as jsf:
        cont = json.load(jsf)
    return tuple(cont["variables"])


class BinningParameters:
    def __init__(self, binningVars):
        self.binningVars = binningVars

    def __call__(self, obj):
        return op.construct(
            "Parameters",
            (op.initList(
                "std::initializer_list<Parameters::value_type::value_type>",
                "Parameters::value_type::value_type", (
                    op.initList(
                        "Parameters::value_type::value_type", "float",
                        (op.extVar("int", "BinningVariable::{}".format(
                            bvNm.replace("ClusEta", "Eta"))), bv(obj)))
                    for bvNm, bv in self.binningVars.items())),)
        )


def getBinningParameters(bVarNames, isElectron=False, moreVars=None, paramDefs=None):
    if isElectron:
        bVarNames = [k.replace("Eta", "ClusEta") for k in bVarNames]
    theDict = dict(paramDefs) if paramDefs else {}
    if moreVars:
        theDict.update(moreVars)
    return BinningParameters({k: theDict[k] for k in bVarNames})


class ScaleFactor:
    def __init__(self, cppDef=None, args=None, iface="ILeptonScaleFactor",
                 systName=None, seedFun=None, defineOnFirstUse=True):
        self._cppDef = cppDef
        self._args = args
        self.sfOp = op.define(iface, cppDef)
        self._systName = systName
        self._seedFun = seedFun
        self.defineOnFirstUse = defineOnFirstUse

    def __call__(self, obj):
        args = list(a(obj) for a in self._args)
        if self._seedFun:
            args.append(self._seedFun(obj))
        expr = self.sfOp.get(*(args + [op.extVar("int", "Nominal")]))
        if self._systName:
            expr = op.systematic(
                expr, name=self._systName,
                up=self.sfOp.get(*(args + [op.extVar("int", "Up")])),
                down=self.sfOp.get(*(args + [op.extVar("int", "Down")])))
        if self.defineOnFirstUse:
            expr = op.defineOnFirstUse(expr)
        return expr


def get_scalefactor(
        objType, key, combine=None, additionalVariables=None, sfLib=None, paramDefs=None,
        lumiPerPeriod=None, periods=None, getFlavour=None, isElectron=False,
        systName=None, seedFun=None, defineOnFirstUse=True):
    """ Construct a scalefactor callable

    .. warning::
        This function is deprecated. Use `correctionlib <https://cms-nanoaod.github.io/correctionlib/>`_ and
        :py:func:`~bamboo.scalefactors.get_correction` instead.

    :param objType: object type: ``"lepton"``, ``"dilepton"``, or ``"jet"``
    :param key: key in ``sfLib`` (or tuple of keys, in case of a nested dictionary),
        or JSON path (or list thereof) if ``sfLib`` is not specified
    :param sfLib: dictionary (or nested dictionary) of scale factors.
        A scale factor entry is either a path to a JSON file, or a list of pairs ``(periods, path)``,
        where ``periods`` is a list of periods found in ``lumiPerPeriod`` and ``path`` is a path
        to the JSON file with the scale factors corresponding to those run periods.
    :param combine: combination strategy for combining different run periods
        (``"weight"`` or ``"sample"``)
    :param paramDefs: dictionary of binning variable definitions (name to callable)
    :param additionalVariables: additional binning variable definitions (TODO: remove)
    :param lumiPerPeriod: alternative definitions and relative weights of run periods
    :param periods: Only combine scale factors for those periods
    :param isElectron: if True, will use supercluster eta instead of eta (for ``"lepton"`` type only)
        (TODO: find a better way of doing that)
    :param systName: name of the associated systematic nuisance parameter
    :param seedFun: (only when combining scalefactor by sampling)
        callable to get a random generator seed for an object, e.g. ``lambda l : l.idx+42``
    :param defineOnFirstUse: wrap with :py:func:`~bamboo.treefunctions.defineOnFirstUse`
        (to define as a column and reuse afterwards),
        this is enabled by default since it is usually more efficient

    :returns: a callable that takes ``(object, variation="Nominal")``
        and returns a floating-point number proxy
    """

    logger.warning("get_scalefactor() is deprecated! Use correctionlib and get_correction() instead")

    #
    # Interpret args, get defaults etc.
    #
    if sfLib is None:
        config = key
    elif isinstance(key, tuple):
        # interpret key=("a", "b") as ...["a"]["b"]
        config = sfLib[key[0]]
        for idx in range(1, len(key)):
            config = config[key[idx]]
    else:
        config = sfLib[key]

    if combine is not None:
        combPrefix = {
            "weight": "W",
            "sample": "Smp"
        }.get(combine, "W")
        if combine == "sample" and not seedFun:
            raise ValueError(
                "If combining by sampling, a seed function needs to be passed to get_scalefactor")

    if getFlavour is None:
        def getFlavour(j):
            return j.hadronFlavour
    getFlavour = partial(lambda getter, j: op.extMethod(
        "IJetScaleFactor::get_flavour")(getter(j)), getFlavour)

    if lumiPerPeriod:
        lumiPerPeriod_default.update(lumiPerPeriod)
    lumiPerPeriod = lumiPerPeriod_default
    if periods is None:
        periods = lumiPerPeriod.keys()

    #
    # Construct scalefactors
    #
    if objType == "lepton":
        iface = "ILeptonScaleFactor"
        if isinstance(config, str):
            return ScaleFactor(
                cppDef=f'const ScaleFactor <<name>>{{"{config}"}};',
                args=(getBinningParameters(
                    getBinningVarNames(config), isElectron=isElectron,
                    moreVars=additionalVariables, paramDefs=paramDefs),),
                iface=iface, systName=systName, defineOnFirstUse=defineOnFirstUse)
        else:
            if combPrefix == "":
                raise ValueError("A combination mode needs to be specified for this scale factor")
            selConfigs = list(
                filter(
                    (lambda elm: elm[0] != 0.),  # only keep those with nonzero lumi
                    ((sum(lumiPerPeriod[ier] for ier in eras if ier in periods), path)
                        for eras, path in config if any(ier in periods for ier in eras))
                ))
            if len(selConfigs) < 1:
                raise RuntimeError(
                    "Zero period configs selected for config {} with periods {}".format(
                        ", ".join(f"({list(eras)} : {path})" for eras, path in config),
                        list(periods)))
            elif len(selConfigs) == 1:
                return ScaleFactor(
                    cppDef=f'const ScaleFactor <<name>>{{"{selConfigs[0][1]}"}};',
                    args=(getBinningParameters(
                        getBinningVarNames(selConfigs[0][1]), isElectron=isElectron,
                        moreVars=additionalVariables, paramDefs=paramDefs),),
                    iface=iface, systName=systName, defineOnFirstUse=defineOnFirstUse)
            else:
                bVarNames = set(chain.from_iterable(getBinningVarNames(iPth)
                                                    for iWgt, iPth in selConfigs))
                return ScaleFactor(
                    cppDef=(
                        'std::unique_ptr<{iface}> tmpSFs_<<name>>[] = {{ {0} }};\n'.format(", ".join(
                            f'std::make_unique<ScaleFactor>("{path}")'
                            for wgt, path in selConfigs), iface=iface)
                        + 'const {cmb}ScaleFactor <<name>>{{ {{ {0} }}, '.format(
                            ", ".join(f"{wgt:e}" for wgt, path in selConfigs), cmb=combPrefix)
                        + (f'std::vector<std::unique_ptr<{iface}>>{{'
                           f'std::make_move_iterator(std::begin(tmpSFs_<<name>>)), '
                           f'std::make_move_iterator(std::end(tmpSFs_<<name>>))}} }};')
                    ),
                    args=(getBinningParameters(
                        bVarNames, isElectron=isElectron,
                        moreVars=additionalVariables, paramDefs=paramDefs),),
                    iface=iface, systName=systName,
                    seedFun=(seedFun if combine == "sample" else None),
                    defineOnFirstUse=defineOnFirstUse
                )
    elif objType == "dilepton":
        iface = "IDiLeptonScaleFactor"
        if isinstance(config, tuple) and len(config) == 4:
            if not all(isinstance(iCfg, str) for iCfg in config):
                raise TypeError(
                    "Config for dilepton scale factor should be quadruplet of paths "
                    f"or list of weights and triplets, found {config}")

            return ScaleFactor(
                cppDef="const DiLeptonFromLegsScaleFactor <<name>>{{{0}}};".format(", ".join(
                    f'std::make_unique<ScaleFactor>("{leplepCfg}")' for leplepCfg in config)),
                args=[(lambda bp: (lambda ll: bp(ll[0])))(
                    getBinningParameters(
                        set(chain(getBinningVarNames(config[0]), getBinningVarNames(config[1]))),
                        moreVars=additionalVariables, paramDefs=paramDefs)),
                      (lambda bp: (lambda ll: bp(ll[1])))(
                          getBinningParameters(
                              set(chain(getBinningVarNames(config[2]), getBinningVarNames(config[3]))),
                              moreVars=additionalVariables, paramDefs=paramDefs))],
                iface=iface, systName=systName, defineOnFirstUse=defineOnFirstUse)
        else:
            raise NotImplementedError("Still to do this part")
    elif objType == "jet":
        iface = "IJetScaleFactor"
        if isinstance(config, tuple) and len(config) == 3:
            if not all(isinstance(iCfg, str) for iCfg in config):
                raise TypeError(
                    f"Config for b-tagging should be triplet of paths "
                    "or list of weights and triplets, found {config}")
            else:
                bVarNames = set(chain.from_iterable(getBinningVarNames(iCfg) for iCfg in config))
                return ScaleFactor(
                    cppDef='const BTaggingScaleFactor <<name>>{{{0}}};'.format(
                        ", ".join(f'"{iCfg}"' for iCfg in config)),
                    args=(getBinningParameters(
                        bVarNames, moreVars=additionalVariables,
                        paramDefs=paramDefs), getFlavour),
                    iface=iface, systName=systName, defineOnFirstUse=defineOnFirstUse)
        else:
            if not (all((isinstance(iCfg, tuple) and len(iCfg) == 3 and all(
                    isinstance(iPth, str) for iPth in iCfg)) for iCfg in config)):
                raise TypeError(
                    "Config for b-tagging should be triplet of paths "
                    f"or list of weights and triplets, found {config}")
            else:
                if combPrefix == "":
                    raise ValueError("A combination mode needs to be specified for this scale factor")
                selConfigs = list(
                    filter(
                        (lambda elm: elm[0] != 0.),  # only keep those with nonzero lumi
                        ((sum(lumiPerPeriod[ier] for ier in eras if ier in periods), path)
                            for eras, path in config if any(ier in periods for ier in eras))))
                if len(selConfigs) < 1:
                    raise RuntimeError("Zero configs")
                elif len(selConfigs) == 1:
                    bVarNames = set(chain.from_iterable(getBinningVarNames(iCfg) for iCfg in selConfigs[0]))
                    return ScaleFactor(
                        cppDef='const BTaggingScaleFactor <<name>>{{{0}}};'.format(
                            ", ".join(f'"{iCfg}"' for iCfg in selConfigs[0])),
                        args=(getBinningParameters(
                            bVarNames, moreVars=additionalVariables, paramDefs=paramDefs), getFlavour),
                        iface=iface, systName=systName, defineOnFirstUse=defineOnFirstUse)
                else:
                    bVarNames = set(chain.from_iterable(getBinningVarNames(iPth)
                                    for iWgt, paths in selConfigs for iPth in paths))
                    return ScaleFactor(
                        cppDef=('std::unique_ptr<{iface}> tmpSFs_<<name>>[] = {{ {0} }};\n'.format(", ".join(
                            'std::make_unique<BTaggingScaleFactor>({})'.format(
                                ", ".join(f'"{iPth}"' for iPth in paths))
                            for wgt, paths in selConfigs), iface=iface)
                            + 'const {cmb}ScaleFactor <<name>>{{ {{ {0} }}, '.format(
                                ", ".join(f"{wgt:e}" for wgt, paths in selConfigs), cmb=combPrefix)
                            + (f'std::vector<std::unique_ptr<{iface}>>{{'
                               f'std::make_move_iterator(std::begin(tmpSFs_<<name>>)), '
                               f'std::make_move_iterator(std::end(tmpSFs_<<name>>))}} }};')
                        ),
                        arg=(getBinningParameters(
                            bVarNames, moreVars=additionalVariables, paramDefs=paramDefs), getFlavour),
                        iface=iface, systName=systName,
                        seedFun=(seedFun if combine == "sample" else None),
                        defineOnFirstUse=defineOnFirstUse)
    else:
        raise ValueError(f"Unknown object type: {objType}")


class BtagSF:
    """ Helper for b- and c-tagging scalefactors using the BTV POG reader """
    def _nano_getPt(jet):
        import bamboo.treeproxies as _tp
        if isinstance(jet._parent, _tp.AltCollectionProxy):
            bs = jet._parent._base
            return bs.brMap["pt"].wrapped.result[jet.idx]  # use nominal always
        else:
            return jet.pt

    def _nano_getEta(jet):
        import bamboo.treefunctions as op
        return op.abs(jet.eta)

    def _nano_getJetFlavour(jet):
        import bamboo.treefunctions as op
        return op.extMethod("BTagEntry::jetFlavourFromHadronFlavour")(jet.hadronFlavour)

    def _translate_btagSFVarToJECVar(btagVarName, prefix="btagSF_"):
        if btagVarName.startswith("up_jes") or btagVarName.startswith("down_jes"):
            if btagVarName.endswith("_jes"):
                return "jesTotal{}".format(btagVarName.split("_jes")[0])
            return "jes{1}{0}".format(*btagVarName.split("_jes"))
        elif btagVarName.startswith("up_") or btagVarName.startswith("down_"):
            tk = btagVarName.split("_")
            return "".join((prefix, "_".join(tk[1:]), tk[0]))
        elif btagVarName in ("up", "down"):
            return "".join((prefix, btagVarName))
        else:
            return btagVarName

    def translateFixedWPCorrelation(btagVarName, prefix="btagSF", year=None):
        if btagVarName.endswith("_correlated"):
            rest = "_".join(btagVarName.split("_")[:-1])
            return f"{prefix}{year}{rest}"
        elif btagVarName.endswith("_uncorrelated"):
            rest = "_".join(btagVarName.split("_")[:-1])
            return f"{prefix}{rest}"
        else:
            return f"{prefix}{btagVarName}"

    def __init__(
            self, taggerName, csvFileName, wp=None, sysType="central", otherSysTypes=None,
            measurementType=None, getters=None, jesTranslate=None, sel=None, uName=None):
        """
        Declare a BTagCalibration (if needed) and BTagCalibrationReader (unique, based on ``uName``),
            and decorate for evaluation

        .. warning::
            This function is deprecated.
            Use `correctionlib <https://cms-nanoaod.github.io/correctionlib/>`_ and
            helpers in :py:mod:`~bamboo.scalefactors` instead.

        :param taggerName: first argument for ``BTagCalibration``
        :param csvFileName: name of the CSV file with scalefactors
        :param wp: working point (used as ``BTagEntry::OP_{wp.upper()}``)
        :param sysType: nominal value systematic type (``"central"``, by default)
        :param otherSysTypes: other systematic types to load in the reader
        :param measurementType: dictionary with measurement type per true flavour (B, C, and UDSG),
            or a string if the same for all (if not specified,
            ``"comb"`` will be used for b- and c-jets, and ``incl`` for light-flavour jets)
        :param getters: dictionary of methods to get the kinematics and classifier for a jet
            (the keys ``Pt``, ``Eta``, ``JetFlavour``, and ``Discri`` are used.
            For the former three, the defaults are for NanoAOD)
        :param jesTranslate: translation function for JEC systematic variations,
            from the names in the CSV file to those used for the jets
            (the default should work for on-the-fly corrections)
        :param sel: a selection in the current graph
        :param uName: unique name, to declare the reader (e.g. sample name)
        """
        logger.warning("BtagSF is deprecated! Use correctionlib and other bamboo helpers instead")

        if otherSysTypes is None:
            otherSysTypes = []
        self.nomName = sysType
        self.varNames = otherSysTypes
        if measurementType is None:  # BTV recommendation for b-tagging with fixed working points
            measurementType = {"B": "comb", "C": "comb", "UDSG": "incl"}
        elif isinstance(measurementType, str):  # if string, use it for all
            measurementType = {fl: measurementType for fl in ("B", "C", "UDSG")}
        calibName = sel._fbe.symbol(
            f'const BTagCalibration <<name>>{{"{taggerName}", "{csvFileName}"}};',
            nameHint=f"bTagCalib_{taggerName}")
        readerName = sel._fbe.symbol(
            'BTagCalibrationReader <<name>>{{BTagEntry::OP_{0}, "{1}", {{ {2} }} }}; // for {3}'.format(
                wp.upper(), sysType, ", ".join(f'"{sv}"' for sv in otherSysTypes), uName),
            nameHint="bTagReader_{}".format("".join(c for c in uName if c.isalnum())))
        from bamboo.root import gbl
        calibHandle = getattr(gbl, calibName)
        readerHandle = getattr(gbl, readerName)
        for flav, measType in measurementType.items():
            readerHandle.load(calibHandle, getattr(gbl.BTagEntry, f"FLAV_{flav}"), measType)
        import bamboo.treefunctions as op
        self.reader = op.extVar("BTagCalibrationReader", readerName)
        if getters is None:
            self.getters = [BtagSF._nano_getJetFlavour, BtagSF._nano_getEta, BtagSF._nano_getPt]
        else:
            self.getters = [
                getters.get("JetFlavour", BtagSF._nano_getJetFlavour),
                getters.get("Eta", BtagSF._nano_getEta),
                getters.get("Pt", BtagSF._nano_getPt)]
            if "Discri" in getters:
                self.getters.append(getters["Discri"])
        self.jesTranslate = jesTranslate if jesTranslate is not None else BtagSF._translate_btagSFVarToJECVar

    def _evalFor(self, var, jet):
        import bamboo.treefunctions as op
        varName_p = op._to.Parameter("std::string", f'"{var}"').result
        return self.reader.eval_auto_bounds(varName_p, *(gtr(jet) for gtr in self.getters))

    def __call__(self, jet, nomVar=None, systVars=None):
        """
        Evaluate the scalefactor for a jet

        Please note that this only gives the applicable scalefactor:
        to obtain the event weight one of the recipes in the
        `POG twiki <https://twiki.cern.ch/twiki/bin/viewauth/CMS/BTagSFMethods>`_
        should be used.

        By default the nominal and systematic variations are taken from
        the :py:class:`bamboo.scalefactors.BtagSF` instance, but they can be
        overriden with the ``nomVar`` and ``systVars`` keyword arguments.
        Please note that when using split uncertainties (e.g. for the reshaping
        method) some uncertainties only apply to specific jet flavours
        (e.g. c-jets) and the csv file contains zeroes for the other flavours.
        Then the user code should check the jet flavours, and call this method
        with the appropriate list of variations for each.
        """
        import bamboo.treefunctions as op
        nom = self._evalFor((nomVar if nomVar is not None else self.nomName), jet)
        if systVars is not None and not systVars:  # empty, no systematic
            return nom
        else:
            if systVars is not None:
                if any(sv not in self.varNames for sv in systVars):
                    logger.error(
                        "B-tag SF systematic variations {} were not loaded, skipping".format(
                            ", ".join(sv for sv in systVars if sv not in self.varNames)))
                    systVars = [sv for sv in systVars if sv in self.varNames]
            else:
                systVars = self.varNames
            return op.systematic(nom, **{self.jesTranslate(var): self._evalFor(var, jet) for var in systVars})


class CorrectionHelper:
    def __init__(
            self, corrOp, inputs, params, systParam=None,
            systNomName="nominal", systVariations=None, defineOnFirstUse=True):
        self.corrOp = corrOp
        self.inputs = inputs
        self.params = params
        self.systParam = systParam
        self.systNomName = systNomName
        self.systVariations = systVariations
        self.defineOnFirstUse = defineOnFirstUse

    @staticmethod
    def makeEvalArg(args):
        return op.initList(
            "std::vector<correction::Variable::Type>",
            "correction::Variable::Type",
            args)

    def __call__(self, obj):
        args = []
        iSyst = None
        for pName in self.inputs:
            if pName == self.systParam:
                iSyst = self.inputs.index(self.systParam)
                args.append(op._tp.makeConst(self.systNomName, "std::string"))
            else:
                p = self.params[pName]
                if callable(p):
                    args.append(p(obj))
                else:
                    args.append(p)
        expr = self.corrOp.evaluate(CorrectionHelper.makeEvalArg(args))
        if self.systVariations:
            variations = {}
            for vName, vCatIndex in self.systVariations.items():
                vArgs = list(args)
                vArgs[iSyst] = op._tp.makeConst(vCatIndex, "std::string")
                vExpr = self.corrOp.evaluate(CorrectionHelper.makeEvalArg(vArgs))
                variations[vName] = vExpr
            expr = op.systematic(expr, **variations)
        if self.defineOnFirstUse:
            expr = op.defineOnFirstUse(expr)
        return expr


_loadedCorrectionSets = {}


class CorrectionInputError(Exception):
    pass


def get_correction(
        path, correction, params=None, systParam=None, systNomName="nominal",
        systVariations=None, systName=None, defineOnFirstUse=True, sel=None):
    """
    Load a correction from a CMS JSON file

    The JSON file is parsed with `correctionlib <https://github.com/cms-nanoAOD/correctionlib>`_.
    The contents and structure of a JSON file can be checked with the
    ``correction`` script, e.g. ``correction summary sf.json``

    :param path: JSON file path
    :param correction: name of the correction inside ``CorrectionSet`` in the JSON file
    :param params: parameter definitions, a dictionary of values or functions
    :param systParam: name of the parameter (category axis) to use for systematic variations
    :param systNomName: name of the systematic category axis to use as nominal
    :param systVariations: systematic variations list or ``{variation: name_in_json}``
    :param systName: systematic uncertainty name (to prepend to names, if 'up' and 'down')
    :param defineOnFirstUse: wrap with :py:func:`~bamboo.treefunctions.defineOnFirstUse`
        (to define as a column and reuse afterwards),
        this is enabled by default since it is usually more efficient
    :param sel: a selection in the current graph (only used to retrieve a pointer to the backend)

    :returns: a callable that takes ``(object)`` and returns the correction
        (with systematic variations, if present and unless a specific variation is requested)
        obtained by evaluating the remaining parameters with the object

    :Example:

    >>> elIDSF = get_correction("EGM_POG_SF_UL.json", "electron_cutbased_looseID",
    >>>     params={"pt": lambda el : el.pt, "eta": lambda el : el.eta},
    >>>     systParam="weight", systNomName="nominal", systName="elID", systVariations=("up", "down")
    >>>     )
    >>> looseEl = op.select(t.Electron, lambda el : el.looseId)
    >>> withDiEl = noSel.refine("withDiEl",
    >>>     cut=(op.rng_len(looseEl) >= 2),
    >>>     weight=[ elIDSF(looseEl[0]), elIDSF(looseEl[1]) ]
    >>>     )
    """
    from .root import loadcorrectionlib
    be = sel._fbe  # backend reference
    loadcorrectionlib(backend=be)
    # TODO real path (but not abs), expanduser?
    if path not in _loadedCorrectionSets:
        _loadedCorrectionSets[path] = (be.symbol(
            f'auto <<name>> = correction::CorrectionSet::from_file("{path}");'), {})
    corrSetName, loadedCorrs = _loadedCorrectionSets[path]
    if correction not in loadedCorrs:
        loadedCorrs[correction] = be.symbol(f'auto <<name>> = {corrSetName}->at("{correction}");')
    corrName = loadedCorrs[correction]
    corrObj = op.extVar("correction::Correction*", corrName)
    from .root import gbl
    corrHandle = getattr(gbl, corrName)
    if systVariations and not isinstance(systVariations, dict):
        systVars = {}
        for vn in systVariations:
            if vn.lower() in ("up", "down") and systName:
                systVars[f"{systName}{vn}"] = vn
            else:
                if systName:
                    logger.warning(
                        "Systematic name given, and variation not up or down, "
                        f"using only variation name '{vn}'")
                systVars[vn] = vn
        systVariations = systVars
    corrInputs = [p.name() for p in corrHandle.inputs()]
    if params is not None:
        missingInputs = set(corrInputs) - set(params.keys())
        if systParam:
            missingInputs.remove(systParam)
        if missingInputs:
            raise CorrectionInputError(
                f"For correction {correction} in {path}, "
                f"the following inputs are missing: {missingInputs}")
        extraInputs = set(params.keys()) - set(corrInputs)
        if extraInputs:
            raise CorrectionInputError(
                f"For correction {correction} in {path}, "
                f"the following inputs are not expected: {extraInputs}")
    return CorrectionHelper(
        corrObj,
        corrInputs,
        params=params,
        systParam=systParam,
        systNomName=systNomName,
        systVariations=systVariations,
        defineOnFirstUse=defineOnFirstUse
    )


def get_bTagSF_fixWP(json_path, tagger, wp, flavour, sel, jet_pt_variation=None, heavy_method="comb",
                     syst_prefix="btagSF_fixWP_", decorr_wps=False, decorr_eras=False, era=None,
                     full_scheme=False, syst_mapping=None, defineOnFirstUse=True):
    """Build correction evaluator for fixed working point b-tagging scale factors

    Loads the b-tagging scale factors as correction object from the JSON file,
    configures the systematic variations, and returns a callable that
    can be evaluated on a jet to return the scale factor.

    :param json_path: JSON file path
    :param tagger: name of the tagger inside the JSON (not the same as in the event!)
    :param wp: working point of the tagger (`"L"`, `"M"`, `"T"`)
    :param flavour: hadron flavour of the jet (`0`, `4`, `5`)
    :param sel: a selection in the current graph (only used to retrieve a pointer to the backend)
    :param jet_pt_variation: if specified, only use that specific systematic variation (e.g. the `nominal`)
        of the jet pt to evaluate the scale factors. By default, the scale factors are evaluated for
        each variation.
    :param heavy_method: B-tagging measurement method for heavy-flavour jets (`"comb"` or `"mujets"`).
        For light jets, there is only the `"incl"` method.
    :param syst_prefix: Prefix to prepend to the name of all resulting the b-tagging systematic variations.
        Variations for light or heavy jets will be prefixed resp.
        by `{syst_prefix}light` or `{syst_prefix}heavy` (unless the full scheme is used).
    :param decorr_wps: If `True`, insert the working point into the systematic name for
        the uncorrelated/statistical component. Otherwise, all working points will be taken as fully
        correlated when using several in the analysis.
    :param decorr_eras: If `True`, use the scale factor uncertainties split into "uncorrelated"
        and "correlated" parts, and insert the `era` name into the variation names.
        If `False`, only use the total scale factor uncertainties (not split).
    :param era: Name of era, used in the name of systematic variations if one of `decorr_eras`
        or `full_scheme` is `True`.
    :param full_scheme: If `True`, use split uncertainty sources as specified in the `full_scheme` argument
    :param syst_mapping: Dictionary used to list the systematics to consider, and
        to map the naming of the full-scheme b-tagging uncertainties
        to variations defined elsewhere in the analysis, for varying them together when needed
        (see example below).
    :param defineOnFirstUse: see description in :py:func:`~bamboo.scalefactors.get_correction`

    :returns: a callable that takes a jet and returns the correction
        (with systematic variations as configured here)
        obtained by evaluating the b-tagging scale factors on the jet

    :Example:

    >>> btvSF_b = get_bTagSF_fixWP("btv.json", "deepJet", "M", 5, sel, syst_prefix="btagSF_",
    >>>                            era="2018UL", full_scheme=True,
    >>>                            syst_mapping={
    >>>                                "pileup": "pileup",
    >>>                                "type3": None,
    >>>                                "jer0": "jer",
    >>>                                "jer1": "jer"
    >>>                             })

    Will result in the following systematic uncertainties:
        - `btagSF_statistic_2018UL{up/down}`: mapped to `{up/down}_statistic` in the JSON
        - `btagSF_pileup{up/down}`: mapped to `{up/down}_pileup` in the JSON, and correlated
            with the `pileup{up/down}` variations in the analysis
        - `btagSF_type3{up/down}`: mapped to `{up/down}_type3` in the JSON
        - `btagSF_jer0{up/down}`: mapped to `{up/down}_jer` in the JSON, and correlated
            with the `jer0{up/down}` variations in the analysis
        - `btagSF_jer1{up/down}`: mapped to `{up/down}_jer` in the JSON, and correlated
            with the `jer1{up/down}` variations in the analysis

    >>> btvSF_b = get_bTagSF_fixWP("btv.json", "deepJet", "M", 5, sel, syst_prefix="btagSF_",
    >>>                            era="2018UL", decorr_wps=True, decorr_eras=True)

    Will result in the following systematic uncertainties:
        - `btagSF_heavy_M_2018UL{up/down}`: mapped to `{up/down}_uncorrelated` in the JSON
        - `btagSF_heavy{up/down}`: mapped to `{up/down}_correlated` in the JSON
    """
    if not era and (decorr_eras or full_scheme):
        raise RuntimeError("Need to specify an era if the nuisances are to be decorrelated across eras, "
                           "or if the full uncertainty scheme is used")
    params = {
        "pt": lambda j: op.forSystematicVariation(j.pt, jet_pt_variation) if jet_pt_variation else j.pt,
        "abseta": lambda j: op.abs(j.eta), "working_point": wp, "flavor": flavour
    }
    systName = syst_prefix + ("light" if flavour == 0 else "heavy")
    wpDecorrTag = f"_{wp}" if decorr_wps else ""
    systVariations = {}
    for d in ("up", "down"):
        if not decorr_eras and not full_scheme:
            systVariations[f"{systName}{d}"] = d
        elif decorr_eras and (not full_scheme or flavour == 0):
            systVariations[f"{systName}{d}"] = f"{d}_correlated"
            systVariations[f"{systName}{wpDecorrTag}_{era}{d}"] = f"{d}_uncorrelated"
        elif full_scheme and flavour > 0:
            systVariations[f"{syst_prefix}statistic{wpDecorrTag}_{era}{d}"] = f"{d}_statistic"
            for var, varBTV in syst_mapping.items():
                if varBTV is None:
                    systVariations[f"{syst_prefix}{var}{d}"] = f"{d}_{var}"
                else:
                    systVariations[f"{var}{d}"] = f"{d}_{varBTV}"

    if tagger == "deepCSV_subjet":
        method = "incl" if flavour == 0 else "lt"
        params["method"] = method
        taggerStr = tagger
    else:
        method = "incl" if flavour == 0 else heavy_method
        taggerStr = f"{tagger}_{method}"

    return get_correction(json_path, taggerStr, params=params,
                          systParam="systematic", systNomName="central",
                          systVariations=systVariations, sel=sel, defineOnFirstUse=defineOnFirstUse)


def makeBtagWeightMeth1a(jets, tagger, wps, workingPoints, sfGetter, effGetters):
    """Construct the full event weight based on b-tagging scale factors
    (fixed working point) and efficiencies

    Combines the b-tagging scale factors and MC efficiencies for fixed working points
    into the event weight needed to correct the simulation (the event weight can then
    directly be passed to a selection). The weight is computed according to
    `Method 1a <https://twiki.cern.ch/twiki/bin/viewauth/CMS/BTagSFMethods#1a_Event_reweighting_using_scale>`_,  # noqa: B950
    with support for several working points.

    While the scale factors can be loaded from the POG-provided JSON files,
    the efficiencies need to be computed by the analyzers.

    :param jets: the jet collection in the event
    :param tagger: the name of the tagger in the event
    :param wps: a list of working points for which the b tagging must be corrected,
        e.g. `["L", "T"]` for computing the weight using scale factors for the `L` and `T`
        working points. Note: always provide the working points in increasing order of "tightness"!
    :param workingPoints: a dictionary providing the working point value for the discriminator
    :param sfGetter: a callable that takes the working point (`str`) and the hadron flavour (`int`)
        and returns the correction object for the b-tagging scale factors of that working point
        and jet flavour (i.e., itself a callable that takes the jet and returns the scale factor).
        See :py:meth:`bamboo.scalefactors.get_bTagSF_fixWP`.
    :param effGetters: a dictionary with keys = working points and values = callables that can be evaluated
        on a jet and return the b-tagging efficiency for that working point. Typically,
        these would be correction objects parameterized using the jet pt, eta and hadron flavour.

    :returns: a weight proxy (with all systematic variations configured in the scale factors)

    :Example:

    >>> btvSF = lambda wp, flav: get_bTagSF_fixWP("btv.json", "deepJet", wp, flav, ...)
    >>> btvEff = {"M": get_correction("my_btag_eff.json", ...)}
    >>> btvWeight = makeBtagWeightMeth1a(tree.jet, "btagDeepFlavB", ["M"], {"M": 0.2783},
    >>>                                  btvSF, btvEff)
    >>> sel_btag = sel.refine("btag", cut=..., weight=btvWeight)
    """
    # Functions selecting the right SF depending on jet flavour
    # The flavour can't be passed as parameter to correctionlib because the
    # uncertainties depend on it.
    bTagSFs = {}
    for wp in wps:
        bTagSFs[wp] = lambda j, wp_=wp: op.multiSwitch(
            (j.hadronFlavour == 5, sfGetter(wp_, 5)(j)),
            (j.hadronFlavour == 4, sfGetter(wp_, 4)(j)),
            sfGetter(wp_, 0)(j))

    wFail = op.extMethod("scalefactorWeightForFailingObject", returnType="double")

    def jet_SF(j):
        factors = []
        # [L,M,T] -> (if discr >= T, then: )
        tightest = wps[-1]
        factors.append((getattr(j, tagger) >= workingPoints[tightest], bTagSFs[tightest](j)))
        # [L,M,T] -> (elif discr >= M, then: ), (elif discr >= L, then: )
        for i in range(len(wps) - 1, 0, -1):
            tighter = wps[i]
            looser = wps[i - 1]
            factors.append((getattr(j, tagger) >= workingPoints[looser],
                            wFail(bTagSFs[tighter](j), effGetters[tighter](j),
                                  bTagSFs[looser](j), effGetters[looser](j))))
        # [L,M,T] -> (else: )
        loosest = wps[0]
        factors.append(wFail(bTagSFs[loosest](j), effGetters[loosest](j)))
        return op.multiSwitch(*factors)

    # method 1a: product over jets, factors depend on discriminator value
    return op.rng_product(jets, jet_SF)


def get_bTagSF_itFit(json_path, tagger_json, tagger_jet, flavour, sel, jet_pt_variation=None,
                     syst_prefix="btagSF_shape_",
                     decorr_eras=False, era=None, syst_mapping=None, defineOnFirstUse=True):
    """Build correction evaluator for continuous (iterativeFit) b-tagging scale factors

    Loads the b-tagging scale factors as correction object from the JSON file,
    configures the systematic variations, and returns a callable that
    can be evaluated on a jet to return the scale factor.

    :param json_path: JSON file path
    :param tagger_json: name of the tagger inside the JSON
    :param tagger_jet: name of the tagger in the tree
    :param flavour: hadron flavour of the jet (`0`, `4`, `5`)
    :param sel: a selection in the current graph (only used to retrieve a pointer to the backend)
    :param jet_pt_variation: see description in :py:func:`~bamboo.scalefactors.get_bTagSF_fixWP`
    :param syst_prefix: Prefix to prepend to the name of all resulting the b-tagging systematic variations.
    :param decorr_eras: If `True`, insert the era into the variation name for statistical uncertainties
    :param era: Name of era, used in the name of systematic variations if `decorr_eras` is `True`
    :param syst_mapping: see description in :py:func:`~bamboo.scalefactors.get_bTagSF_fixWP`,
        with the difference that here the "basic" (non-JES-related) variations are already included
        no matter what.
    :param defineOnFirstUse: see description in :py:func:`~bamboo.scalefactors.get_correction`

    :returns: a callable that takes a jet and returns the correction
        (with systematic variations as configured here)
        obtained by evaluating the b-tagging scale factors on the jet

    :Example:

    >>> btvSF_b = get_bTagSF_itFit("btv.json", "deepJet", "btagDeepFlavB", 5, sel, syst_prefix="btagSF_",
    >>>                            decorr_eras=True, era="2018UL",
    >>>                            syst_mapping={"jesTotal": "jes"})

    Will result in the following systematic uncertainties:
        - `btagSF_hfstats1_2018UL{up/down}`: mapped to `{up/down}_hfstats1` in the JSON
        - `btagSF_hfstats2_2018UL{up/down}`: mapped to `{up/down}_hfstats2` in the JSON
        - `btagSF_lfstats1_2018UL{up/down}`: mapped to `{up/down}_lfstats1` in the JSON
        - `btagSF_lfstats2_2018UL{up/down}`: mapped to `{up/down}_lfstats2` in the JSON
        - `btagSF_hf{up/down}`: mapped to `{up/down}_hf` in the JSON
        - `btagSF_lf{up/down}`: mapped to `{up/down}_lf` in the JSON
        - `btagSF_jesTotal{up/down}`: mapped to `{up/down}_jes` in the JSON, and correlated
            with the `jesTotal{up/down}` variations in the analysis

    (for c jets, the `hf` and `lf` variations are absent and replaced by `cferr1` and `cferr2`)
    """

    systListUnCorr = []
    if flavour == 4:
        systListCorr = ["cferr1", "cferr2"]
    else:
        systListCorr = ["hf", "lf"]
        statSyst = ["hfstats1", "hfstats2", "lfstats1", "lfstats2"]
        if decorr_eras:
            systListUnCorr = statSyst
        else:
            systListCorr += statSyst

    systVariations = {}
    for d in ("up", "down"):
        for var in systListCorr:
            systVariations[f"{syst_prefix}{var}{d}"] = f"{d}_{var}"
        for var in systListUnCorr:
            systVariations[f"{syst_prefix}{var}_{era}{d}"] = f"{d}_{var}"
        if syst_mapping is not None:
            for var, varBTV in syst_mapping.items():
                systVariations[f"{var}{d}"] = f"{d}_{varBTV}"

    params = {
        "pt": lambda j: op.forSystematicVariation(j.pt, jet_pt_variation) if jet_pt_variation else j.pt,
        "abseta": lambda j: op.abs(j.eta),
        "discriminant": lambda j: getattr(j, tagger_jet), "flavor": flavour
    }
    return get_correction(json_path, f"{tagger_json}_shape", params=params,
                          systParam="systematic", systNomName="central",
                          systVariations=systVariations, sel=sel, defineOnFirstUse=defineOnFirstUse)


def makeBtagWeightItFit(jets, sfGetter):
    """Construct the full event weight based on b-tagging scale factors (continous/iterativeFit)

    Combines the b-tagging scale factors into the event weight needed to correct the simulation
    (the event weight can then directly be passed to a selection),
    by making a product of the scale factors over all jets.
    See the `note <https://twiki.cern.ch/twiki/bin/view/CMS/BTagShapeCalibration>`_ about
    correcting the normalization when using these scale factors.

    :param jets: the jet collection in the event
    :param sfGetter: a callable that takes the the hadron flavour (`int`)
        and returns the correction object for the b-tagging scale factors of that jet flavour
        (i.e., itself a callable that takes the jet and returns the scale factor)
        See :py:meth:`bamboo.scalefactors.get_bTagSF_itFit`.

    :returns: a weight proxy (with all systematic variations configured in the scale factors)

    :Example:

    >>> btvSF = lambda flav: get_bTagSF_itFit("btv.json", "deepJet", "btagDeepFlavB", flav, ...)
    >>> btvWeight = makeBtagWeightItFit(tree.jet, btvSF)
    >>> sel_btag = sel.refine("btag", cut=..., weight=btvWeight)
    """
    # The flavour can't be passed as parameter to correctionlib because the
    # uncertainties depend on it.
    def bTagSF(j):
        return op.multiSwitch(
            (j.hadronFlavour == 5, sfGetter(5)(j)),
            (j.hadronFlavour == 4, sfGetter(4)(j)),
            sfGetter(0)(j))
    return op.rng_product(jets, bTagSF)



import math
import os.path

import pytest

from bamboo.analysisutils import (fatjet_gmrValues_tau21DDT,
                                  fatjet_gmsValues_tau21DDT, fatjet_jmrValues,
                                  fatjet_jmsValues, fatjet_puppi_msd_params)
from CMSJMECalculators.jetdatabasecache import JetDatabaseCache

testData = os.path.join(os.path.dirname(__file__), "data")

jecDBCache = JetDatabaseCache("JECDatabase", repository="cms-jet/JECDatabase",
                              cachedir=testData, mayWrite=True)
jrDBCache = JetDatabaseCache("JRDatabase", repository="cms-jet/JRDatabase",
                             cachedir=testData, mayWrite=True)


def toRVecFloat(values):
    from bamboo.root import gbl
    RVec_float = gbl.ROOT.VecOps.RVec["float"]
    res = RVec_float(len(values), 0.)
    for i, val in enumerate(values):
        res[i] = val
    return res


def toRVecInt(values):
    from bamboo.root import gbl
    RVec_int = gbl.ROOT.VecOps.RVec["int"]
    res = RVec_int(len(values), 0)
    for i, val in enumerate(values):
        res[i] = val
    return res


def getEventWith(f, condition=lambda ev: True, treeName="Events"):
    tup = f.Get(treeName)
    tup.GetEntry(0)
    i = 0
    while not condition(tup):
        i += 1
        tup.GetEntry(i)
    yield tup


def getJetMETArgs(tup, isMC=True, forMET=False, isMETFixEE2017=False, addHEM2018Issue=False):
    args = [
        toRVecFloat(tup.Jet_pt),
        toRVecFloat(tup.Jet_eta),
        toRVecFloat(tup.Jet_phi),
        toRVecFloat(tup.Jet_mass),
        toRVecFloat(tup.Jet_rawFactor),
        toRVecFloat(tup.Jet_area),
    ]
    if forMET:
        args += [
            toRVecFloat(tup.Jet_muonSubtrFactor),
            toRVecFloat(tup.Jet_neEmEF),
            toRVecFloat(tup.Jet_chEmEF),
        ]
    if not (forMET and isMETFixEE2017):
        args.append(toRVecInt(tup.Jet_jetId if addHEM2018Issue else []))
    args.append(tup.fixedGridRhoFastjetAll)
    if isMC:
        args += [
            toRVecInt(tup.Jet_partonFlavour),
            ((tup.run << 20) + (tup.luminosityBlock << 10) + tup.event
             + 1 + (int(tup.Jet_eta[0] / .01) if tup.nJet != 0 else 0)),
            toRVecFloat(tup.GenJet_pt),
            toRVecFloat(tup.GenJet_eta),
            toRVecFloat(tup.GenJet_phi),
            toRVecFloat(tup.GenJet_mass)
        ]
    else:
        args += [toRVecInt([]), 0, toRVecFloat([]), toRVecFloat([]), toRVecFloat([]), toRVecFloat([])]
    if forMET:
        args += [tup.RawMET_phi, tup.RawMET_pt]
        if not isMETFixEE2017:
            args += [tup.MET_MetUnclustEnUpDeltaX, tup.MET_MetUnclustEnUpDeltaY]
        else:
            args += [tup.METFixEE2017_MetUnclustEnUpDeltaX, tup.METFixEE2017_MetUnclustEnUpDeltaY]
        args += [toRVecFloat(getattr(tup, f"CorrT1METJet_{varNm}"))
                 for varNm in ("rawPt", "eta", "phi", "area", "muonSubtrFactor")]
        args += [toRVecFloat([]), toRVecFloat([])]
        if isMETFixEE2017:
            args += [tup.MET_phi, tup.MET_pt, tup.METFixEE2017_phi, tup.METFixEE2017_pt]
    return args


def getFatJetArgs(tup, isMC=True, addHEM2018Issue=False):
    args = [
        toRVecFloat(tup.FatJet_pt),
        toRVecFloat(tup.FatJet_eta),
        toRVecFloat(tup.FatJet_phi),
        toRVecFloat(tup.FatJet_mass),
        toRVecFloat(tup.FatJet_rawFactor),
        toRVecFloat(tup.FatJet_area),
        toRVecFloat(tup.FatJet_msoftdrop),
        toRVecInt(tup.FatJet_subJetIdx1),
        toRVecInt(tup.FatJet_subJetIdx2),
        toRVecFloat(tup.SubJet_pt),
        toRVecFloat(tup.SubJet_eta),
        toRVecFloat(tup.SubJet_phi),
        toRVecFloat(tup.SubJet_mass),
        toRVecInt(tup.FatJet_jetId if addHEM2018Issue else []),
        tup.fixedGridRhoFastjetAll
    ]
    if isMC:
        args += [
            ((tup.run << 20) + (tup.luminosityBlock << 10) + tup.event
             + 1 + (int(tup.Jet_eta[0] / .01) if tup.nJet != 0 else 0)),
            toRVecFloat(tup.GenJetAK8_pt),
            toRVecFloat(tup.GenJetAK8_eta),
            toRVecFloat(tup.GenJetAK8_phi),
            toRVecFloat(tup.GenJetAK8_mass),
            toRVecFloat(tup.SubGenJetAK8_pt),
            toRVecFloat(tup.SubGenJetAK8_eta),
            toRVecFloat(tup.SubGenJetAK8_phi),
            toRVecFloat(tup.SubGenJetAK8_mass)
        ]
    else:
        args += [0, toRVecFloat([]), toRVecFloat([]), toRVecFloat([]), toRVecFloat([]),
                 toRVecFloat([]), toRVecFloat([]), toRVecFloat([]), toRVecFloat([])]
    return args


def configureCalc(
        config, jecTag=None, jerTag=None, jetType="AK4PFchs",
        levels=None, levels_l1=None, splitJER=False, uncSources=None):
    if jecTag:
        if levels:
            config.jec = [jecDBCache.getPayload(jecTag, iLev, jetType) for iLev in levels]
        if uncSources:
            config.jesUncertainties = [
                (src, jecDBCache.getPayload(jecTag, "UncertaintySources", jetType))
                for src in uncSources]
        if levels_l1:  # for MET
            config.l1Jec = [jecDBCache.getPayload(jecTag, iLev, jetType) for iLev in levels_l1]
    if jerTag:
        config.ptResolution = jrDBCache.getPayload(jerTag, "PtResolution", jetType)
        config.ptResolutionSF = jrDBCache.getPayload(jerTag, "SF", jetType)
        config.splitJER = splitJER
        config.useGenMatch = True


def configureFatJetCalc(
        config, isMC=False, year=None, doSmearing=False,
        jmr=None, jms=None, isTau21DDT=False, gmr=None, gms=None, puppiGen=None,
        puppiRecoCorrCen=None, puppiRecoCorrFwd=None, puppiResolCen=None, puppiResolFwd=None):
    # defaults
    if isMC:  # mass scale
        if jms is None:
            jms = fatjet_jmsValues[year]
        if gms is None:
            if isTau21DDT:
                gms = fatjet_gmsValues_tau21DDT[year]
            else:
                gms = jms
    if jms is not None:
        print(f"Setting JMS values {jms}")
        config.jms = jms
    if gms is not None:
        config.gms = gms
    if doSmearing:  # mass resolutions
        if jmr is None:
            jmr = fatjet_jmrValues[year]
        if gmr is None:
            if isTau21DDT:
                gmr = fatjet_gmrValues_tau21DDT[year]
            else:
                gmr = jmr
        print(f"Setting JMR values: {jmr}")
        config.jmr = jmr
        print(f"Setting GMR values: {gmr}")
        config.gmr = gmr
    # PUPPI: always
    if puppiGen is None:
        puppiGen = fatjet_puppi_msd_params["gen"]
    if puppiRecoCorrCen is None:
        puppiRecoCorrCen = fatjet_puppi_msd_params["reco_cen"]
    if puppiRecoCorrFwd is None:
        puppiRecoCorrFwd = fatjet_puppi_msd_params["reco_fwd"]
    if puppiResolCen is None:
        puppiResolCen = fatjet_puppi_msd_params["resol_cen"]
    if puppiResolFwd is None:
        puppiResolFwd = fatjet_puppi_msd_params["resol_fwd"]

    config.puppiGenFormula = puppiGen
    config.puppi_reco_cen = puppiRecoCorrCen
    config.puppi_reco_for = puppiRecoCorrFwd
    config.puppi_resol_cen = puppiResolCen
    config.puppi_resol_for = puppiResolFwd


@pytest.fixture(scope="module")
def nanojetargsMC16():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016.root"))
    for tup in getEventWith(f, (lambda tup_: tup_.nJet >= 5)):
        yield getJetMETArgs(tup, isMC=True, forMET=False)


@pytest.fixture(scope="module")
def nanojetargsMC16_postvalues():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_JMEKin_bTagShape_puWeight.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        jet_vars = {"nominal": (toRVecFloat(tup.Jet_pt_nom), toRVecFloat(tup.Jet_mass_nom))}
        from itertools import chain
        jet_vars.update(dict(chain.from_iterable(
            {f"jer{i:d}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"Jet_{ivar}_jer{i:d}{vdir}")) for ivar in ("pt", "mass"))
                for vdir in ("Up", "Down")}.items() for i in range(6))))
        jet_vars.update(dict(chain.from_iterable(
            {f"jes{src}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"Jet_{ivar}_jes{src}{vdir}".format(ivar, src)))
                for ivar in ("pt", "mass"))
                for vdir in ("Up", "Down")}.items() for src in ("AbsoluteStat", "AbsoluteScale")
        )))
        res.append((getJetMETArgs(tup, isMC=True, forMET=False), jet_vars))
    yield res


@pytest.fixture(scope="module")
def nanoMETargsMC16_postvalues():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_JMEKin_bTagShape_puWeight.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        met_vars = {
            "nominal": (tup.MET_T1_pt, tup.MET_T1_phi),
            "unclustEnup": (tup.MET_T1_pt_unclustEnUp, tup.MET_T1_phi_unclustEnUp),
            "unclustEndown": (tup.MET_T1_pt_unclustEnDown, tup.MET_T1_phi_unclustEnDown)
        }
        met_vars.update({
            f"{nm}{var.lower()}": (getattr(tup, f"MET_T1_pt_{nm}{var}"),
                                   getattr(tup, f"MET_T1_phi_{nm}{var}"))
            for var in ("Up", "Down") for nm in (
                [f"jer{i:d}" for i in range(6)]
                + [f"jes{jsnm}" for jsnm in ("AbsoluteScale", "AbsoluteStat")])
        })
        res.append((tuple(getJetMETArgs(tup, isMC=True, forMET=True)), met_vars))
    yield res


@pytest.fixture(scope="module")
def nanoSmearMETargsMC16_postvalues():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_JMEKin_bTagShape_puWeight.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        met_vars = {
            "nominal": (tup.MET_T1Smear_pt, tup.MET_T1Smear_phi),
            "unclustEnup": (tup.MET_T1Smear_pt_unclustEnUp, tup.MET_T1Smear_phi_unclustEnUp),
            "unclustEndown": (tup.MET_T1Smear_pt_unclustEnDown, tup.MET_T1Smear_phi_unclustEnDown)
        }
        met_vars.update({
            f"{nm}{var.lower()}": (getattr(tup, f"MET_T1Smear_pt_{nm}{var}"),
                                   getattr(tup, f"MET_T1Smear_phi_{nm}{var}"))
            for var in ("Up", "Down") for nm in (
                [f"jer{i:d}" for i in range(6)]
                + [f"jes{jsnm}" for jsnm in ("AbsoluteScale", "AbsoluteStat")])
        })
        res.append((tuple(getJetMETArgs(tup, isMC=True, forMET=True)), met_vars))
    yield res


@pytest.fixture(scope="module")
def nanoMETFixEE2017argsMC17_postvalues():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2017postproc_JMEKin_METFixEE2017.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        met_vars = {
            "nominal": (tup.METFixEE2017_T1_pt, tup.METFixEE2017_T1_phi),
            "unclustEnup": (tup.METFixEE2017_pt_unclustEnUp, tup.METFixEE2017_phi_unclustEnUp),
            "unclustEndown": (tup.METFixEE2017_pt_unclustEnDown, tup.METFixEE2017_phi_unclustEnDown)
        }
        met_vars.update({
            f"{nm}{var.lower()}": (getattr(tup, f"METFixEE2017_T1_pt_{nm}{var}"),
                                   getattr(tup, f"METFixEE2017_T1_phi_{nm}{var}"))
            for var in ("Up", "Down") for nm in (
                ["jer"] + [f"jes{jsnm}" for jsnm in ("AbsoluteScale", "AbsoluteStat")])
        })
        res.append((tuple(getJetMETArgs(tup, isMC=True, forMET=True, isMETFixEE2017=True)), met_vars))
    yield res


@pytest.fixture(scope="module")
def nanojetargsMC18_postvalues_hem():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2018postproc_hemfix.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        jet_vars = {"nominal": (toRVecFloat(tup.Jet_pt_nom), toRVecFloat(tup.Jet_mass_nom))}
        from itertools import chain
        jet_vars.update(dict(chain.from_iterable(
            {f"{src}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"Jet_{ivar}_{src}{vdir}".format(ivar, src)))
                for ivar in ("pt", "mass"))
                for vdir in ("Up", "Down")}.items() for src in (
                    "jer", "jesAbsoluteStat", "jesAbsoluteScale", "jesHEMIssue")
        )))
        res.append((getJetMETArgs(tup, isMC=True, forMET=False, addHEM2018Issue=True), jet_vars))
    yield res


@pytest.fixture(scope="module")
def nanofatjetargsMC16_postvalues():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_FatJetKin.root"))
    tup = f.Get("Events")
    res = []
    for i in range(tup.GetEntries()):
        tup.GetEntry(i)
        jet_vars = {
            "nominal": (toRVecFloat(tup.FatJet_pt_nom),
                        toRVecFloat(tup.FatJet_mass_nom),
                        toRVecFloat(tup.FatJet_msoftdrop_nom)),
        }
        from itertools import chain
        jet_vars.update(dict(chain.from_iterable(
            {f"jer{i:d}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"FatJet_{ivar}_jer{i:d}{vdir}"))
                for ivar in ("pt", "mass", "msoftdrop"))
                for vdir in ("Up", "Down")}.items() for i in range(6))))
        jet_vars.update(dict(chain.from_iterable(
            {f"jes{src}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"FatJet_{ivar}_jes{src}{vdir}".format(ivar, src)))
                for ivar in ("pt", "mass", "msoftdrop"))
                for vdir in ("Up", "Down")}.items() for src in ("AbsoluteStat", "AbsoluteScale")
        )))
        jet_vars.update(dict(chain.from_iterable(
            {f"{src}{vdir.lower()}": tuple(
                toRVecFloat(getattr(tup, f"FatJet_{ivar}_{src}{vdir}".format(ivar, src)))
                for ivar in ("mass", "msoftdrop"))
                for vdir in ("Up", "Down")}.items() for src in ("jmr", "jms")
        )))
        res.append((getFatJetArgs(tup, isMC=True), jet_vars))
    yield res


@pytest.fixture(scope="module")
def nanofatjetargsMC16():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_FatJetKin.root"))
    for tup in getEventWith(f):
        yield getFatJetArgs(tup, isMC=True, addHEM2018Issue=False)


@pytest.fixture(scope="module")
def jetvarcalc_empty():
    from CMSJMECalculators.config import JetVariations
    config = JetVariations()
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def jetvarcalcMC16_smear():
    from CMSJMECalculators.config import JetVariations
    config = JetVariations()
    configureCalc(config, jerTag="Summer16_25nsV1_MC", splitJER=True)
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def jetvarcalcMC16_jec():
    from CMSJMECalculators.config import JetVariations
    config = JetVariations()
    configureCalc(config, jerTag="Summer16_25nsV1_MC", splitJER=True,
                  jecTag="Summer16_07Aug2017_V11_MC", levels=["L1FastJet", "L2Relative"])
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def jetvarcalcMC16_jesunc():
    from CMSJMECalculators.config import JetVariations
    config = JetVariations()
    configureCalc(config, jerTag="Summer16_25nsV1_MC", splitJER=True,
                  jecTag="Summer16_07Aug2017_V11_MC", levels=["L1FastJet", "L2Relative"],
                  uncSources=["AbsoluteStat", "AbsoluteScale"])
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def metvarcalcMC16_jesunc():
    from CMSJMECalculators.config import METVariations
    config = METVariations()
    configureCalc(config, jerTag="Summer16_25nsV1_MC", splitJER=True,
                  jecTag="Summer16_07Aug2017_V11_MC", levels=["L1FastJet", "L2Relative"],
                  uncSources=["AbsoluteStat", "AbsoluteScale"], levels_l1=["L1FastJet"])
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def metvarcalcMC17_FixEE():
    from CMSJMECalculators.config import FixEE2017METVariations
    # a better test even would be to have a different JEC than production
    config = FixEE2017METVariations()
    configureCalc(config, jerTag="Fall17_V3_MC", splitJER=False,
                  jecTag="Fall17_17Nov2017_V32_MC", levels=["L1FastJet", "L2Relative"],
                  uncSources=["AbsoluteStat", "AbsoluteScale"], levels_l1=["L1FastJet"])
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def jetvarcalcMC18_hem():
    from CMSJMECalculators.config import JetVariations
    config = JetVariations()
    configureCalc(config, jerTag="Autumn18_V7b_MC", splitJER=False,
                  jecTag="Autumn18_V19_MC", levels=["L1FastJet", "L2Relative"],
                  uncSources=["AbsoluteStat", "AbsoluteScale"])
    config.addHEM2018Issue = True
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


@pytest.fixture(scope="module")
def fatjetvarcalcMC16():
    from CMSJMECalculators.config import FatJetVariations
    config = FatJetVariations()
    configureCalc(config, jetType="AK8PFPuppi", jerTag="Summer16_25nsV1_MC", splitJER=True,
                  jecTag="Summer16_07Aug2017_V11_MC", levels=["L1FastJet", "L2Relative"],
                  uncSources=["AbsoluteStat", "AbsoluteScale"])
    configureFatJetCalc(config, isMC=True, year="2016", doSmearing=True, isTau21DDT=False)
    from bamboo.root import loadJMESystematicsCalculators
    loadJMESystematicsCalculators()
    yield config.create()


def test_jetvarcalc_empty(jetvarcalc_empty):
    assert jetvarcalc_empty


def test_jetvarcalcMC16_smear(jetvarcalcMC16_smear):
    assert jetvarcalcMC16_smear


def test_jetvarcalcMC16_nano_smear(jetvarcalcMC16_smear, nanojetargsMC16):
    res = jetvarcalcMC16_smear.produce(*nanojetargsMC16)
    assert res


def test_jetvarcalcMC16_nano_jec(jetvarcalcMC16_jec, nanojetargsMC16):
    res = jetvarcalcMC16_jec.produce(*nanojetargsMC16)
    assert res


def test_jetvarcalcMC16_nano_jesunc(jetvarcalcMC16_jesunc, nanojetargsMC16):
    res = jetvarcalcMC16_jesunc.produce(*nanojetargsMC16)
    assert res


def test_fatjetvarcalcMC16_nano(fatjetvarcalcMC16, nanofatjetargsMC16):
    res = fatjetvarcalcMC16.produce(*nanofatjetargsMC16)
    assert res


def isclose_float(a, b, tol=1.):
    from bamboo.root import gbl
    return math.isclose(a, b, rel_tol=tol * gbl.std.numeric_limits["float"].epsilon())


def compareJets(names, calcRes, postValues, tol=1.):
    hasDiff = False
    for ky, (post_pt, post_mass) in postValues.items():
        idx = names.index(ky)
        print(ky, "pt", calcRes.pt(idx), post_pt)
        print(ky, "m ", calcRes.mass(idx), post_mass)
        if not (all(isclose_float(a, b, tol=tol)
                    for a, b in zip(post_pt, calcRes.pt(idx)))
                and all(isclose_float(a, b, tol=tol)
                        for a, b in zip(post_mass, calcRes.mass(idx)))):
            print(f"FAIL: Difference for {ky}")
            hasDiff = True
    return not hasDiff


def compareFatJets(namesAll, namesM, calcRes, postValues, tol=1.):
    hasDiff = False
    for ky, post in postValues.items():
        if ky in namesAll:
            idx = namesAll.index(ky)
            (post_pt, post_mass, post_msd) = post
            print(ky, "pt ", calcRes.pt(idx), post_pt)
            print(ky, "m  ", calcRes.mass(idx), post_mass)
            print(ky, "msd", calcRes.msoftdrop(idx), post_msd)
            eq_pt = all(isclose_float(a, b, tol=tol) for a, b in zip(post_pt, calcRes.pt(idx)))
            eq_m = all(isclose_float(a, b, tol=tol) for a, b in zip(post_mass, calcRes.mass(idx)))
            eq_msd = all(isclose_float(a, b, tol=tol) for a, b in zip(post_msd, calcRes.msoftdrop(idx)))
            if not (eq_pt and eq_m and eq_msd):
                what = ", ".join((["pt"] if not eq_pt else [])
                                 + (["mass"] if not eq_m else [])
                                 + (["msd"] if not eq_msd else []))
                print(f"FAIL: Difference for {ky} (in {what})")
                hasDiff = True
        else:
            idx = len(namesAll) + namesM.index(ky)
            (post_mass, post_msd) = post
            print(ky, "m  ", calcRes.mass(idx), post_mass)
            print(ky, "msd", calcRes.msoftdrop(idx), post_msd)
            eq_m = all(isclose_float(a, b, tol=tol) for a, b in zip(post_mass, calcRes.mass(idx)))
            eq_msd = all(isclose_float(a, b, tol=tol) for a, b in zip(post_msd, calcRes.msoftdrop(idx)))
            if not (eq_m and eq_msd):
                what = ", ".join((["mass"] if not eq_m else [])
                                 + (["msd"] if not eq_msd else []))
                print(f"FAIL: Difference for {ky} (in {what})")
                hasDiff = True
    return not hasDiff


def compareMET(names, calcRes, postValues, reltol_pt=1.e-6, reltol_phi=1.e-6, abstol_phi=1.e-6):
    hasDiff = False
    for ky, (post_pt, post_phi) in postValues.items():
        idx = names.index(ky)
        print(ky, "pt ", calcRes.pt(idx), post_pt)
        print(ky, "phi", calcRes.phi(idx), post_phi)
        if not (math.isclose(calcRes.pt(idx), post_pt, rel_tol=1.e-6)
                and math.isclose(calcRes.phi(idx), post_phi, rel_tol=1.e-6, abs_tol=1.e-6)):
            print(f"FAIL: Difference for {ky}")
            hasDiff = True
    return not hasDiff


def test_jetvarcalc_nanopost_jesunc(jetvarcalcMC16_jesunc, nanojetargsMC16_postvalues):
    for nanojetargsMC16, postValues in nanojetargsMC16_postvalues:
        assert compareJets([str(nm) for nm in jetvarcalcMC16_jesunc.available()],
                           jetvarcalcMC16_jesunc.produce(*nanojetargsMC16),
                           postValues)


def test_metvarcalc_nanopost_jesunc(metvarcalcMC16_jesunc, nanoMETargsMC16_postvalues):
    metvarcalcMC16_jesunc.setIsT1SmearedMET(False)
    for nanoMETargsMC16, postValues in nanoMETargsMC16_postvalues:
        assert compareMET([str(nm) for nm in metvarcalcMC16_jesunc.available()],
                          metvarcalcMC16_jesunc.produce(*nanoMETargsMC16),
                          postValues)


def test_metvarcalc_nanopost_jesunc_T1Smear(
        metvarcalcMC16_jesunc,
        nanoSmearMETargsMC16_postvalues):
    metvarcalcMC16_jesunc.setIsT1SmearedMET(True)
    for nanoMETargsMC16, postValues in nanoSmearMETargsMC16_postvalues:
        assert compareMET([str(nm) for nm in metvarcalcMC16_jesunc.available()],
                          metvarcalcMC16_jesunc.produce(*nanoMETargsMC16), postValues)


def test_metvarcalc_nanopost_jesunc_MCFixEE2017(
        metvarcalcMC17_FixEE,
        nanoMETFixEE2017argsMC17_postvalues):
    metvarcalcMC17_FixEE.setIsT1SmearedMET(False)
    for nanoMETargsMC17FixEE, postValues in nanoMETFixEE2017argsMC17_postvalues:
        assert compareMET([str(nm) for nm in metvarcalcMC17_FixEE.available()],
                          metvarcalcMC17_FixEE.produce(*nanoMETargsMC17FixEE), postValues,
                          reltol_pt=2.e-6, reltol_phi=2.e-6, abstol_phi=2.e-6)


def test_jetvarcalc_nanopost_jesunc_HEM(jetvarcalcMC18_hem, nanojetargsMC18_postvalues_hem):
    for nanojetargsMC18, postValues in nanojetargsMC18_postvalues_hem:
        assert compareJets([str(nm) for nm in jetvarcalcMC18_hem.available()],
                           jetvarcalcMC18_hem.produce(*nanojetargsMC18),
                           postValues, tol=2.)


def test_fatjetvarcalc_nanopost_jesunc(fatjetvarcalcMC16, nanofatjetargsMC16_postvalues):
    for fatjetargs, postValues in nanofatjetargsMC16_postvalues:
        avlAll = [str(nm) for nm in fatjetvarcalcMC16.available()]
        avlM = [str(nm) for nm in fatjetvarcalcMC16.available("mass") if nm not in avlAll]
        assert compareFatJets(avlAll, avlM,
                              fatjetvarcalcMC16.produce(*fatjetargs),
                              postValues)
